/*
** eptest.c - Utility to read and write the configuration EEPROM on an
**            ADM-XRC series card.
**
** Warning: do not use this application to write to the EEPROM unless
**          you know what you are doing!
**
** Utility for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2000-2013 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>

#if defined(_WIN32)

/* Windows */
#include <windows.h>
#include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _tscanf_s(buffer, ...) scanf(buffer, __VA_ARGS__)

#endif

#include <admxrc2.h>

#include <sdk_common.h>

/*
** Return TRUE if 'loc' is an EEPROM location containing
** adapter configuration information that could render the
** card unusable if incorrectly programmed.
*/
static bool_t
protected_location(
	unsigned long loc,
	ADMXRC2_BOARD_TYPE	board)
{
	switch (board) {
	case ADMXRC2_BOARD_ADMXRC:
	case ADMXRC2_BOARD_ADMXRC_P:
	case ADMXRC2_BOARD_ADMXRC2_LITE:
		return (loc < 0x2BU);

	case ADMXRC2_BOARD_ADMXRC2:
	case ADMXRC2_BOARD_ADPWRC2:
	case ADMXRC2_BOARD_ADPDRC2:
	case ADMXRC2_BOARD_ADMXRC4LS:
	case ADMXRC2_BOARD_ADMXRC4LX:
	case ADMXRC2_BOARD_ADMXRC4SX:
	case ADMXRC2_BOARD_ADMPCIE6S1:
		return (loc < 0x2DU);

	case ADMXRC2_BOARD_ADMXPL:
	case ADMXRC2_BOARD_ADMXP:
	case ADMXRC2_BOARD_ADPXPI:
	case ADMXRC2_BOARD_ADMXRC4FX:
	case ADMXRC2_BOARD_ADPEXRC4FX:
	case ADMXRC2_BOARD_ADMXRC5LX:
	case ADMXRC2_BOARD_ADMXRC5T1:
	case ADMXRC2_BOARD_ADMXRC5T2:
	case ADMXRC2_BOARD_ADCPXRC4LX:
	case ADMXRC2_BOARD_ADMAMC5A2:
	case ADMXRC2_BOARD_ADMXRC5TZ:
	case ADMXRC2_BOARD_ADMXRC5T2ADV:
	case ADMXRC2_BOARD_ADMXRC5T2ADV6:
	case ADMXRC2_BOARD_ADMXRC5TDA1:
	case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
		return FALSE;

	default:
		return TRUE;
	}

	/* Don't allow programming if board type is invalid */
	return TRUE;
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
	int ret = 0;
	ADMXRC2_HANDLE hCard = ADMXRC2_HANDLE_INVALID_VALUE;
	ADMXRC2_STATUS status;
	ADMXRC2_CARD_INFO	cardInfo;
	Option options[4];
	Arguments arguments;
	OptionType paramType[2];
	TCHAR* paramInfo[2];
	TCHAR* paramMeaning[2];
	int width;

	/* Set up expected options */
	options[0].key = _T("card");
	options[0].type = Option_uint;
	options[0].def.uintVal = 0;
	options[0].meaning = _T("ID of card to use");
	options[1].key = _T("index");
	options[1].type = Option_uint;
	options[1].def.uintVal = 0;
	options[1].meaning = _T("index of card to use");
	options[2].key = _T("f");
	options[2].type = Option_boolean;
	options[2].def.booleanVal = FALSE;
	options[2].meaning = _T("If TRUE, don't confirm when writing");
	options[3].key = _T("unlock");
	options[3].type = Option_boolean;
	options[3].def.booleanVal = FALSE;
	options[3].meaning = _T("If TRUE, allow write to protected locations");
	arguments.nOption = sizeof(options) / sizeof(Option);
	arguments.option = options;

	paramType[0] = Option_uint;
	paramInfo[0] = _T("location");
	paramMeaning[0] = _T("EEPROM location");
	paramType[1] = Option_uint;
	paramInfo[1] = _T("data");
	paramMeaning[1] = _T("Data value to write");
	arguments.nParamType = sizeof(paramType) / sizeof(OptionType);
	arguments.minNParam = 0;
	arguments.paramType = paramType;
	arguments.paramInfo = paramInfo;
	arguments.paramMeaning = paramMeaning;

	parse_command_line(_T("EPTEST utility"), _T("eptest"), argc, argv, &arguments);

	if (options[0].specified && options[1].specified) {
		_tprintf(_T("Do not specify both -card and -index\n"));
		ret = -1;
		goto done;
	}

	if (!options[1].specified) {
		ADMXRC2_CARDID cardID = options[0].value.uintVal;

		status = ADMXRC2_OpenCard(cardID, &hCard);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with ID %lu: %s\n"), (unsigned long)cardID, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
		unsigned int index = options[1].value.uintVal;

		status = ADMXRC2_OpenCardByIndex(index, &hCard);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	}

    status = ADMXRC2_GetCardInfo(hCard, &cardInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to get card info: %s\n"),	ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	switch (cardInfo.BoardType) {
	case ADMXRC2_BOARD_ADMXRC:
	case ADMXRC2_BOARD_ADMXRC_P:
	case ADMXRC2_BOARD_ADMXRC2_LITE:
	case ADMXRC2_BOARD_ADMXRC2:
	case ADMXRC2_BOARD_ADPWRC2:
	case ADMXRC2_BOARD_ADPDRC2:
	case ADMXRC2_BOARD_ADMXRC4LS:
	case ADMXRC2_BOARD_ADMXRC4LX:
	case ADMXRC2_BOARD_ADMXRC4SX:
	case ADMXRC2_BOARD_ADMPCIE6S1:
		width = 2;
		break;

	case ADMXRC2_BOARD_ADMXPL:
	case ADMXRC2_BOARD_ADMXP:
	case ADMXRC2_BOARD_ADPXPI:
	case ADMXRC2_BOARD_ADMXRC4FX:
	case ADMXRC2_BOARD_ADPEXRC4FX:
	case ADMXRC2_BOARD_ADMXRC5LX:
	case ADMXRC2_BOARD_ADMXRC5T1:
	case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADCPXRC4LX:
	case ADMXRC2_BOARD_ADMAMC5A2:
	case ADMXRC2_BOARD_ADMXRC5TZ:
	case ADMXRC2_BOARD_ADMXRC5T2ADV:
	case ADMXRC2_BOARD_ADMXRC5T2ADV6:
	case ADMXRC2_BOARD_ADMXRC5TDA1:
	case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
		width = 4;
		break;

	default:
		_tprintf(_T("This utility must be run on one of the following cards:\n\n"));
		_tprintf(_T("\tADM-XRC\n"));
		_tprintf(_T("\tADM-XRC-P\n"));
		_tprintf(_T("\tADM-XRC-II-Lite\n"));
		_tprintf(_T("\tADM-XRC-II\n"));
		_tprintf(_T("\tADM-XPL\n"));
		_tprintf(_T("\tADM-XP\n"));
		_tprintf(_T("\tADP-WRC-II\n"));
		_tprintf(_T("\tADP-DRC-II\n"));
		_tprintf(_T("\tADP-XPI\n"));
		_tprintf(_T("\tADM-XRC-4LX\n"));
		_tprintf(_T("\tADM-XRC-4SX\n"));
		_tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
		_tprintf(_T("\tADPE-XRC-4FX\n"));
		_tprintf(_T("\tADM-XRC-5LX\n"));
		_tprintf(_T("\tADM-XRC-5T1\n"));
		_tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
		_tprintf(_T("\tADM-XRC-5TZ\n"));
		_tprintf(_T("\tADM-XRC-5T-DA1\n"));
		_tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
		_tprintf(_T("\tADM-PCIE-6S1\n"));
		ret = -1;
		goto done;
	}

	if (arguments.nParam < 1) {
		/* Dump all of configuration ROM */
		unsigned long	index;
		uint32_t data;

		_tprintf(_T("Selected card ID is %lu(0x%lX)\n\n"),
			(unsigned long)cardInfo.CardID, (unsigned long)cardInfo.CardID);
		if (width == 2) {
			for (index = 0; index < 0x20; index++) {
				ADMXRC2_ReadConfig(hCard, index, &data);
				_tprintf(_T("   [0x%02lX] = 0x%04lX\t\t"), index, (unsigned long)data);
				ADMXRC2_ReadConfig(hCard, index + 0x20, &data);
				_tprintf(_T("[0x%02lX] = 0x%04lX\n"), index + 0x20, (unsigned long)data);
			}
		} else {
			for (index = 0; index < 0x20; index++) {
				ADMXRC2_ReadConfig(hCard, index, &data);
				_tprintf(_T("   [0x%02lX] = 0x%08lX\t\t"), index, (unsigned long)data);
				ADMXRC2_ReadConfig(hCard, index + 0x20, &data);
				_tprintf(_T("[0x%02lX] = 0x%08lX\n"), index + 0x20, (unsigned long)data);
			}
		}
	} else if (arguments.nParam < 2) {
		/* Display specified location of configuration ROM */
		uint32_t data;

    _tprintf(_T("Selected card ID is %lu(0x%lX)\n"),
      (unsigned long) cardInfo.CardID, (unsigned long) cardInfo.CardID);
		ADMXRC2_ReadConfig(hCard, arguments.param[0].uintVal, &data);
		if (width == 2) {
			_tprintf(_T("[0x%0lX] = 0x%04lX\n"), (unsigned long)arguments.param[0].uintVal, (unsigned long)data);
		} else {
			_tprintf(_T("[0x%0lX] = 0x%08lX\n"), (unsigned long)arguments.param[0].uintVal, (unsigned long)data);
		}
	} else {
		/* Write to specified location of configuration ROM */
		TCHAR c;
		int res;

		if (!options[2].value.booleanVal) {
      _tprintf(_T("Selected card ID is %lu(0x%lX)\n"),
        (unsigned long) cardInfo.CardID, (unsigned long) cardInfo.CardID);
      _tprintf(_T("Warning: this will write the value %lu/0x%lX to EEPROM location 0x%lX\n"),
				(unsigned long)arguments.param[1].uintVal,
				(unsigned long)arguments.param[1].uintVal,
				(unsigned long)arguments.param[0].uintVal);
			_tprintf(_T("Are you sure you want to continue (y/n)? "));
			res = _tscanf_s(_T("%c"), &c);
			if (!res || res == EOF) {
				c = _T('n');
			}
		} else {
			c = _T('y');
		}

		if (protected_location(arguments.param[0].uintVal, cardInfo.BoardType) && !options[3].value.booleanVal) {
			_tprintf(_T("Permission denied to protected location.\n"));
		} else {
			if (c == _T('y') || c == _T('Y')) {
				ADMXRC2_WriteConfig(hCard, (unsigned long) arguments.param[0].uintVal, arguments.param[1].uintVal);
			} else {
				_tprintf(_T("Aborted.\n"));
			}
		}
	}

done:
	if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
		ADMXRC2_CloseCard(hCard);
	}

	return ret;
}


