/*
** flash.c - utility to program, erase and verify the Flash memory on
**           an ADM-XRC series card.
**
** Warning: Although this utility performs checks in order to ensure
**          that you do not attempt to program the Flash memory with a
**          bitstream for the wrong type of FPGA, ensure that you do
**          not program the flash with an invalid bitstream. Doing so
**          and powering up the FPGA for an extended period of time
**          can damage devices on the card.
**
** Utility for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2003-2013 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#if defined(_WIN32)

/* Windows */
#include <windows.h>
#include <tchar.h>

#else

#include <strings.h>

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _tscanf_s(buffer, ...) scanf(buffer, __VA_ARGS__)
# if defined(__VXWORKS__) || defined(__vxworks)
#   include <ctype.h>
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif

#endif

#include <admxrc2.h>
#include <admxrc2/flash.h>

#include <sdk_common.h>
#include <platform_endian.h>

#if !defined(_MSC_VER)
/* Emulate _tfopen_s ("secure file open") from MSVC++ */
static int
_tfopen_s(
  FILE** ppFile,
  const TCHAR* pFilename,
  const TCHAR* pMode)
{
  FILE* pFile;

  pFile = fopen(pFilename, pMode);
  if (NULL == pFile) {
    return -1;
  }
  *ppFile = pFile;
  return 0;
}
#endif

#if defined(__VXWORKS__) || defined(__vxworks)
static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}
#endif

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  ADMXRC2_HANDLE hCard = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_FLASH_INFO flashInfo;
	int ret = 0;
  const double clockFreq = 33.0e6;
	ADMXRC2_STATUS status;
	ADMXRC2_CARD_INFO cardInfo;
	ADMXRC2_IMAGE pImage = NULL;
	unsigned long imageSize;
	uint8_t* pFlash = NULL;
	TCHAR* command;
	Arguments arguments;
	Option options[3];
	OptionType paramType[4];
	TCHAR* paramInfo[4];
	TCHAR* paramMeaning[4];
	bool_t doChkblank, doErase, doProgram, doReadback, doVerify, doFill;
	unsigned long bootFlagAddr;
	unsigned long imageAddr;
	unsigned long eraseAddr;
  unsigned long eraseLimit;
	int nError = 0;
	bool_t b32BitBootFlag;
  bool_t bFailsafeImage;
  uint8_t fillValue = 0x55U;

	options[0].key = _T("card");
	options[0].type = Option_uint;
	options[0].def.uintVal = 0;
	options[0].meaning = _T("ID of card to use");
	options[1].key = _T("index");
	options[1].type = Option_uint;
	options[1].def.uintVal = 0;
	options[1].meaning = _T("index of card to use");
	options[2].key = _T("failsafe");
	options[2].type = Option_boolean;
	options[2].def.booleanVal = FALSE;
	options[2].meaning = _T("program failsafe image");
	arguments.nOption = 3;
	arguments.option = options;

	paramType[0] = Option_string;
	paramInfo[0] = _T("command");
	paramMeaning[0] = _T("One of erase, program, readback, verify or chkblank");
	paramType[1] = Option_path;
	paramInfo[1] = _T("file");
	paramMeaning[1] = _T("Bitstream filename (program or verify commands only)");
	paramType[2] = Option_uint;
	paramInfo[2] = _T("offset");
	paramMeaning[2] = _T("Offset for readback (readback command only)");
	paramType[3] = Option_uint;
	paramInfo[3] = _T("length");
	paramMeaning[3] = _T("Length for readback (readback command only)");
	arguments.nParamType = sizeof(paramType) / sizeof(OptionType);
	arguments.minNParam = 1;
	arguments.paramType = paramType;
	arguments.paramInfo = paramInfo;
	arguments.paramMeaning = paramMeaning;

	parse_command_line(_T("FLASH utility"), _T("flash"), argc, argv, &arguments);

	command = arguments.param[0].stringVal;
	doProgram = _tcsicmp(command, _T("program")) == 0;
	doVerify = _tcsicmp(command, _T("verify")) == 0;
	doErase = _tcsicmp(command, _T("erase")) == 0;
	doChkblank = _tcsicmp(command, _T("chkblank")) == 0;
	doReadback = _tcsicmp(command, _T("readback")) == 0;
	doFill = _tcsicmp(command, _T("fill55")) == 0;
	if (!doProgram && !doVerify && !doErase && !doChkblank && !doReadback && !doFill) {
		_tprintf(_T("*** Please specify a valid command. Use flash /? to get help.\n"));
		ret = -1;
		goto done;
	}

	if (doProgram || doVerify) {
		if (arguments.nParam < 2) {
			_tprintf(_T("*** Must specify a filename for programming or verification\n"));
			ret = -1;
			goto done;
		}
	}

	if (doReadback) {
		if (arguments.nParam < 4) {
			_tprintf(_T("*** Must specify a filename, offset and length for readback\n"));
			ret = -1;
			goto done;
		}
	}

	if (options[0].specified && options[1].specified) {
		_tprintf(_T("*** Do not specify both -card and -index\n"));
		ret = -1;
		goto done;
	}

	if (!options[1].specified) {
		ADMXRC2_CARDID cardID = options[0].value.uintVal;

		status = ADMXRC2_OpenCard(cardID, &hCard);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("*** Failed to open card with ID %lu: %s\n"),
				(unsigned long)cardID, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
		unsigned int index = options[1].value.uintVal;

		status = ADMXRC2_OpenCardByIndex(index, &hCard);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("*** Failed to open card with index %lu: %s\n"),
				(unsigned long)index, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	}

  bFailsafeImage = options[2].value.booleanVal;

  ADMXRC2_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

	switch (cardInfo.BoardType) {
	case ADMXRC2_BOARD_ADMXRC:
	case ADMXRC2_BOARD_ADMXRC_P:
	case ADMXRC2_BOARD_ADMXRC2:
	case ADMXRC2_BOARD_ADMPCIE6S1:
	case ADMXRC2_BOARD_ADMXPL:
	case ADMXRC2_BOARD_ADMXP:
	case ADMXRC2_BOARD_ADPWRC2:
	case ADMXRC2_BOARD_ADPDRC2:
	case ADMXRC2_BOARD_ADPXPI:
	case ADMXRC2_BOARD_ADMXRC4LS:
	case ADMXRC2_BOARD_ADMXRC4LX:
	case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADPEXRC4FX:
    if (bFailsafeImage) {
      _tprintf(_T("*** +failsafe option is not supported for this model.\n"));
      ret = -1;
      goto done;
    }
    break;

	case ADMXRC2_BOARD_ADMXRC4FX:
	case ADMXRC2_BOARD_ADMXRC5LX:
	case ADMXRC2_BOARD_ADMXRC5T1:
	case ADMXRC2_BOARD_ADMXRC5T2:
	case ADMXRC2_BOARD_ADMAMC5A2:
	case ADMXRC2_BOARD_ADMXRC5TZ:
	case ADMXRC2_BOARD_ADMXRC5T2ADV:
	case ADMXRC2_BOARD_ADMXRC5T2ADV6:
	case ADMXRC2_BOARD_ADMXRC5TDA1:
	case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
		break;

	default:
		_tprintf(_T("*** This utility must be run on one of the following cards:\n\n"));
		_tprintf(_T("\tADM-XRC\n"));
		_tprintf(_T("\tADM-XRC-P\n"));
		_tprintf(_T("\tADM-XRC-II\n"));
		_tprintf(_T("\tADM-XPL\n"));
		_tprintf(_T("\tADM-XP\n"));
		_tprintf(_T("\tADP-WRC-II\n"));
		_tprintf(_T("\tADP-DRC-II\n"));
		_tprintf(_T("\tADP-XPI\n"));
		_tprintf(_T("\tADM-XRC-4LX\n"));
		_tprintf(_T("\tADM-XRC-4SX\n"));
		_tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
		_tprintf(_T("\tADPE-XRC-4FX\n"));
		_tprintf(_T("\tADM-XRC-5LX\n"));
		_tprintf(_T("\tADM-XRC-5T1\n"));
		_tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
		_tprintf(_T("\tADM-XRC-5TZ\n"));
		_tprintf(_T("\tADM-XRC-5T-DA1\n"));
		_tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
		_tprintf(_T("\tADM-PCIE-6S1\n"));
		ret = -1;
		goto done;
	}

	status = ADMXRC2_GetFlashInfo(hCard, &flashInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("*** Failed to get Flash information: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

  switch (cardInfo.BoardType) {
  case ADMXRC2_BOARD_ADMXRC:
  case ADMXRC2_BOARD_ADMXRC_P:
  case ADMXRC2_BOARD_ADMXRC2:
  case ADMXRC2_BOARD_ADPWRC2:
  case ADMXRC2_BOARD_ADPDRC2:
    bootFlagAddr = flashInfo.UseableStart + 0x8000U;
    imageAddr = flashInfo.UseableStart + 0x8001U;
    eraseAddr = flashInfo.UseableStart;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    b32BitBootFlag = FALSE;
    break;

  case ADMXRC2_BOARD_ADMPCIE6S1:
    bootFlagAddr = flashInfo.UseableStart + 0x8000U;
    imageAddr = flashInfo.UseableStart + 0x8001U;
    eraseAddr = flashInfo.UseableStart;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    b32BitBootFlag = FALSE;
    break;

  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    bootFlagAddr = flashInfo.UseableStart + 0x20000U;
    imageAddr = flashInfo.UseableStart + 0x20004U;
    eraseAddr = flashInfo.UseableStart + 0x20000U;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    b32BitBootFlag = TRUE;
    break;

  case ADMXRC2_BOARD_ADMXPL:
  case ADMXRC2_BOARD_ADMXP:
  case ADMXRC2_BOARD_ADPXPI:
  case ADMXRC2_BOARD_ADPEXRC4FX:
    bootFlagAddr = flashInfo.UseableStart;
    imageAddr = flashInfo.UseableStart + 0x20000U;
    eraseAddr = flashInfo.UseableStart;
    eraseLimit = flashInfo.UseableStart + flashInfo.UseableLength;
    b32BitBootFlag = TRUE;
    break;

  case ADMXRC2_BOARD_ADMXRC4FX:
  case ADMXRC2_BOARD_ADMXRC5LX:
  case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5TZ:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5T2ADV6:
  case ADMXRC2_BOARD_ADMXRC5TDA1:
  case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
    if (bFailsafeImage) {
      bootFlagAddr = flashInfo.UseableStart + 0xC00000U;
      imageAddr = flashInfo.UseableStart + 0xC20000U;
      eraseAddr = flashInfo.UseableStart + 0xC00000U;
      eraseLimit = flashInfo.UseableStart + 0x1800000U;
    } else {
      bootFlagAddr = flashInfo.UseableStart;
      imageAddr = flashInfo.UseableStart + 0x20000U;
      eraseAddr = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0xC00000U;
    }
    b32BitBootFlag = TRUE;
    break;

  case ADMXRC2_BOARD_ADMAMC5A2:
    if (bFailsafeImage) {
      bootFlagAddr = flashInfo.UseableStart + 0x1000000U;
      imageAddr = flashInfo.UseableStart + 0x1020000U;
      eraseAddr = flashInfo.UseableStart + 0x1000000U;
      eraseLimit = flashInfo.UseableStart + 0x2000000U;
    } else {
      bootFlagAddr = flashInfo.UseableStart;
      imageAddr = flashInfo.UseableStart + 0x20000U;
      eraseAddr = flashInfo.UseableStart;
      eraseLimit = flashInfo.UseableStart + 0x1000000U;
    }
    b32BitBootFlag = TRUE;
    break;

  default:
    assert(FALSE);
    break;
  }

	if (doProgram || doVerify) {
		status = ADMXRC2_LoadBitstream(hCard, arguments.param[1].pathVal, &pImage, &imageSize);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("*** Failed to load bitstream '%s': %s\n"),
				arguments.param[1].pathVal, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
    imageSize = flashInfo.UseableLength;
		pImage = (ADMXRC2_IMAGE)malloc(imageSize);
		if (pImage == NULL) {
			_tprintf(_T("*** Failed to allocate 0x%lX bytes for image\n"), (unsigned long)flashInfo.UseableLength);
			ret = -1;
			goto done;
		}
		memset(pImage, 0xff, imageSize);
	}

  _tprintf(_T("Flash type is %s, %lu(0x%lX) kiB\n"),
    flashInfo.DeviceName, (unsigned long)flashInfo.DeviceSize / 1024, (unsigned long)flashInfo.DeviceSize / 1024);
  _tprintf(_T("Useable region is 0x%lX-0x%lX\n"),
    (unsigned long)flashInfo.UseableStart, (unsigned long)flashInfo.UseableStart + flashInfo.UseableLength - 1);
  if (doProgram || doVerify) {
    _tprintf(_T("Bitstream length is %lu(0x%lX) bytes\n"), imageSize, imageSize);
  }

  if (cardInfo.NumClock > 0) {
	  status = ADMXRC2_SetClockRate(hCard, ADMXRC2_CLOCK_LCLK, clockFreq, NULL);
	  if (status != ADMXRC2_SUCCESS) {
		  _tprintf(_T("Failed to set LCLK to %.1fMHz: %s\n"),
			  clockFreq / 1.0e6, ADMXRC2_GetStatusString(status));
		  ret = -1;
		  goto done;
	  }
  }
	
	if (doErase || doProgram) {
		unsigned long i;
		ADMXRC2_FLASHBLOCK_INFO blockInfo;

		_tprintf(_T("Erasing Flash from 0x%lX-0x%lX...\n"), eraseAddr, eraseLimit - 1);
		i = eraseAddr;
		while (i < eraseLimit) {
			status = ADMXRC2_GetFlashBlockInfo(hCard, i, &blockInfo);
			if (status != ADMXRC2_SUCCESS) {
				_tprintf(_T("*** Failed to get Flash block info for location 0x%lX: %s\n"),
					(unsigned long)i, ADMXRC2_GetStatusString(status));
				nError++;
				break;
			}

			_tprintf(_T("0x%lX\r"), i);
      fflush(stdout);

			status = ADMXRC2_EraseFlash(hCard, blockInfo.Start, blockInfo.Length);
			if (status != ADMXRC2_SUCCESS) {
				_tprintf(_T("*** Failed to erase Flash at 0x%lX-0x%lX: %s\n"),
					(unsigned long)blockInfo.Start,
					(unsigned long)blockInfo.Start + blockInfo.Length - 1,
					ADMXRC2_GetStatusString(status));
				nError++;
				break;
			}

			i += blockInfo.Length;
		}

		if (nError) {
			_tprintf(_T("*** Erase failed - aborting.\n"));
			ret = -1;
			goto done;
		}
	}
		
	if (doProgram) {
		uint8_t* pData;
		unsigned long i, remaining, chunk;
		ADMXRC2_FLASHBLOCK_INFO	blockInfo;

		/* Write the bitstream into the flash */
		if (!nError) {
			_tprintf(_T("Writing bitstream image at 0x%lX-0x%lX...\n"), imageAddr, imageAddr + imageSize - 1);
			i = imageAddr;
			remaining = imageSize;
			pData = (uint8_t*)pImage;
			while (remaining) {
				status = ADMXRC2_GetFlashBlockInfo(hCard, i, &blockInfo);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to get Flash block info for location 0x%lX: %s\n"),
						(unsigned long)i, ADMXRC2_GetStatusString(status));
					nError++;
					break;
				}

				chunk = blockInfo.Length - (i - blockInfo.Start);
				if (chunk > remaining) {
					chunk = remaining;
				}
				_tprintf(_T("0x%lX\r"), i);
        fflush(stdout);

				status = ADMXRC2_WriteFlash(hCard, i, chunk, pData);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to write Flash at 0x%lX-0x%lX: %s\n"),
						i, i + chunk - 1, ADMXRC2_GetStatusString(status));
					nError++;
					break;
				}

				i += chunk;
				remaining -= chunk;
				pData += chunk;
			}
		}

		if (nError) {
			_tprintf(_T("*** Programming failed - aborting\n"));
			ret = -1;
			goto done;
		}
	}
	
	if (doFill) {
		uint8_t* pData;
		unsigned long i, remaining, chunk;
		ADMXRC2_FLASHBLOCK_INFO	blockInfo;

		/* Write the bitstream into the flash */
		if (!nError) {
			_tprintf(_T("Filling Flash at 0x%lX-0x%lX with 0x%02X...\n"),
        (unsigned long)flashInfo.UseableStart, (unsigned long)flashInfo.UseableLength, (unsigned int)fillValue);
      i = flashInfo.UseableStart;
      remaining = flashInfo.UseableLength - i;
			pData = (uint8_t*)pImage;
      memset(pImage, (int)fillValue, imageSize);
			while (remaining) {
				status = ADMXRC2_GetFlashBlockInfo(hCard, i, &blockInfo);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to get Flash block info for location 0x%lX: %s\n"),
						(unsigned long)i, ADMXRC2_GetStatusString(status));
					nError++;
					break;
				}

				chunk = blockInfo.Length - (i - blockInfo.Start);
        if (chunk > imageSize) {
          chunk = imageSize;
        }
				if (chunk > remaining) {
					chunk = remaining;
				}
				_tprintf(_T("0x%lX\r"), i);
        fflush(stdout);

				status = ADMXRC2_WriteFlash(hCard, i, chunk, pData);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to write Flash at 0x%lX-0x%lX: %s\n"),
						i, i + chunk - 1, ADMXRC2_GetStatusString(status));
					nError++;
					break;
				}

				i += chunk;
				remaining -= chunk;
			}
		}

		if (nError) {
			_tprintf(_T("*** Filling failed - aborting\n"));
			ret = -1;
			goto done;
		}
	}
	
	if (!doReadback) {
		ADMXRC2_FLASHBLOCK_INFO	blockInfo;
		unsigned long verifyStart;
		unsigned long verifyLimit;
		unsigned long i, j, remaining;
		unsigned long pos;
		unsigned long chunk;
		uint8_t* pBuf;
		uint8_t* pTmp = (uint8_t*)pImage;

		if (doProgram || doVerify) {
			verifyStart = imageAddr;
			verifyLimit = imageAddr + imageSize;
			_tprintf(_T("Verifying data from 0x%lX-0x%lX...\n"), verifyStart, verifyLimit - 1);
		} else {
			verifyStart = eraseAddr;
			verifyLimit = eraseLimit;
			_tprintf(_T("Blank checking from 0x%lX-0x%lX...\n"), verifyStart, verifyLimit - 1);
		}

    remaining = verifyLimit - verifyStart;
		pFlash = (uint8_t*)malloc(remaining);
		if (pFlash == NULL) {
			_tprintf(_T("*** Failed to allocate memory to read back flash\n"));
			ret = -1;
			goto done;
		}

    pBuf = pFlash;
    pos = verifyStart;
    i = 0;
    remaining = verifyLimit - verifyStart;
		while (pos < verifyLimit) {
			status = ADMXRC2_GetFlashBlockInfo(hCard, pos, &blockInfo);
			if (status != ADMXRC2_SUCCESS) {
				_tprintf(_T("*** Failed to get Flash block info for location 0x%lX: %s\n"),
					(unsigned long)i, ADMXRC2_GetStatusString(status));
				nError++;
				break;
			}

			chunk = blockInfo.Length - (pos - blockInfo.Start);
			if (chunk > remaining) {
				chunk = remaining;
			}
			_tprintf(_T("0x%lX\r"), pos);
      fflush(stdout);

			status = ADMXRC2_ReadFlash(hCard, pos, chunk, pBuf);
			if (status != ADMXRC2_SUCCESS) {
				_tprintf(_T("*** Failed to read back Flash memory at 0x%lX-0x%lX: %s\n"),
					i, i + chunk - 1, ADMXRC2_GetStatusString(status));
				nError++;
				break;
			}

			for (j = 0; j < chunk; j++) {
				if (pTmp[i + j] != pBuf[j]) {
					if (nError < 10) {
						_tprintf(_T("*** Error at Flash location 0x%lX: expected=0x%02lX, got=0x%02lX\n"),
							(unsigned long)(pos + j), (unsigned long)pTmp[i + j], (unsigned long)pBuf[j]);
					} else if (nError == 10) {
						_tprintf(_T("*** (more errors)\n"));
					}
					nError++;
				}
			}

#if 0
			if (nError) {
				break;
			}
#endif

      i += chunk;
      remaining -= chunk;
      pos += chunk;
      pBuf += chunk;
    }

		if (nError) {
			_tprintf(_T("*** Flash %s failed with %lu data error(s)\n"),
				doChkblank ? _T("blank check") : _T("verification"), (unsigned long) nError);
			goto done;
		} else {
			if (doProgram || doVerify) {
				_tprintf(_T("Flash is correct - bitstream verified.\n"));
			} else {
				_tprintf(_T("Flash is correct - no bitstream present.\n"));
			}
		}

		if (doProgram) {
			if (b32BitBootFlag) {
				uint32_t val32;

				/*
				** Set the flag that causes the FPGA to be loaded from flash,
				** and bitstream length.
				*/
				val32 = (imageSize - 1) << 8;
				_tprintf(_T("Setting load-on-reset flag at 0x%lX-0x%lX to 0x%08lX...\n"),
					(unsigned long)bootFlagAddr,
					(unsigned long)bootFlagAddr + sizeof(val32) - 1,
					(unsigned long)val32);
				status = ADMXRC2_WriteFlash(hCard, bootFlagAddr, sizeof(val32), &val32);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to set load-on-reset flag: %s\n"), ADMXRC2_GetStatusString(status));
					nError++;
				}
			} else {
				uint8_t val8;

				/* Set the flag that causes the FPGA to be loaded from flash */
				val8 = 0x0U;
				_tprintf(_T("Setting load-on-powerup flag at 0x%lX to 0x%02lX...\n"),
					(unsigned long)bootFlagAddr, (unsigned long)val8);
				status = ADMXRC2_WriteFlash(hCard, bootFlagAddr, 0x1U, &val8);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to set load-on-reset flag: %s\n"), ADMXRC2_GetStatusString(status));
					nError++;
				}
			}
		}

		if ((doProgram || doVerify) && !nError) {
			if (b32BitBootFlag) {
				uint32_t expected, bootFlag;

				expected = (imageSize - 1) << 8;
				status = ADMXRC2_ReadFlash(hCard, bootFlagAddr, sizeof(bootFlag), &bootFlag);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to read load-on-reset flag: %s\n"), ADMXRC2_GetStatusString(status));
					nError++;
				} else {
					if (bootFlag != expected) {
						_tprintf(_T("*** Load-on-reset flag is 0x%08lx, should be 0x%08lx\n"), (unsigned long)bootFlag, (unsigned long)expected);
						nError++;
					}
				}
			} else {
				uint8_t expected, bootFlag;

				expected = 0x0U;
				status = ADMXRC2_ReadFlash(hCard, bootFlagAddr, 0x1U, &bootFlag);
				if (status != ADMXRC2_SUCCESS) {
					_tprintf(_T("*** Failed to read back load-on-powerup flag: %s\n"), ADMXRC2_GetStatusString(status));
					nError++;
				} else {
					if (bootFlag != 0x0U) {
						_tprintf(_T("*** Load-on-powerup flag is 0x%2lx, should be 0x%2lx\n"), (unsigned long)bootFlag, (unsigned long)expected);
						nError++;
					}
				}
			}

			if (nError) {
				_tprintf(_T("*** The bitstream may not be loaded on power up.\n"));
				goto done;
			} else {
				_tprintf(_T("The bitstream will be loaded on power up.\n"));
			}
		}
	} else {
		ADMXRC2_FLASHBLOCK_INFO blockInfo;
		unsigned long readStart;
		unsigned long readLength;
		unsigned long i, remaining;
		unsigned long pos;
		unsigned long chunk;
		size_t n;
		uint8_t* pBuf;
		FILE* pFile = NULL;

		readStart = arguments.param[2].uintVal;
		readLength = arguments.param[3].uintVal;
		_tprintf(_T("Reading data from 0x%lX-0x%lX...\n"), readStart, readStart + readLength - 1);

		pFlash = (uint8_t*)malloc(readLength);
		if (pFlash == NULL) {
			_tprintf(_T("*** Failed to allocate memory to read back flash\n"));
			ret = -1;
			goto done;
		}

		pBuf = pFlash;
		remaining = readLength;
		pos = readStart;
		i = 0;
		while (i < readLength) {
			status = ADMXRC2_GetFlashBlockInfo(hCard, pos, &blockInfo);
			if (status != ADMXRC2_SUCCESS) {
				_tprintf(_T("*** Failed to get Flash block info for location 0x%lX: %s\n"), (unsigned long)i, ADMXRC2_GetStatusString(status));
				nError++;
				break;
			}

			chunk = blockInfo.Length - (pos - blockInfo.Start);
			if (chunk > remaining) {
				chunk = remaining;
			}
			_tprintf(_T("0x%lX\r"), pos);
      fflush(stdout);

			status = ADMXRC2_ReadFlash(hCard, pos, chunk, pBuf);
			if (status != ADMXRC2_SUCCESS) {
				_tprintf(_T("*** Failed to read back Flash memory at 0x%lX-0x%lX: %s\n"), i, i + chunk - 1, ADMXRC2_GetStatusString(status));
				nError++;
				break;
			}

			if (nError) {
				break;
			}

			i += chunk;
			remaining -= chunk;
			pos += chunk;
			pBuf += chunk;
		}

    if (_tfopen_s(&pFile, arguments.param[1].pathVal, _T("wb"))) {
			ret = -1;
			_tprintf(_T("*** Failed to open file '%s' for writing\n"), arguments.param[1].pathVal);
			goto done;
		}
		n = fwrite(pFlash, 1, readLength, pFile);
		if (n != readLength) {
			ret = -1;
			_tprintf(_T("*** Failed to write all data for file '%s' (%lu vs %lu)\n"),
				arguments.param[1].pathVal, (long)n, (long)readLength);
			goto done;
		}
		fflush(pFile);
		fclose(pFile);
	}

done:
	if (pFlash != NULL) {
		free(pFlash);
	}

	if (pImage != NULL) {
		if (doProgram || doVerify) {
			ADMXRC2_UnloadBitstream(pImage);
		} else {
			free(pImage);
		}
	}
	
	if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
		ADMXRC2_CloseCard(hCard);
	}
   
	return ret;
}
