/*
** info.c - displays information about an ADM-XRC series card.
**
** Utility for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2000-2014 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#if defined(_WIN32)

/* Windows */
#include <windows.h>
#include <tchar.h>

#else

#include <strings.h>

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf
# define _stprintf_s snprintf
# define _tcscpy_s(pDst, n, pSrc) do { strncpy(pDst, pSrc, n); (pDst)[(n) - 1] = _T('\0'); } while (0)
# define _tcscat_s(pDst, n, pSrc) do { strncat(pDst, pSrc, n); (pDst)[(n) - 1] = _T('\0'); } while (0)

#endif

#include <admxrc2.h>

#include <sdk_common.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static TCHAR modelNames[][32] = {
  _T("ADM-XRC"),
  _T("ADM-XRC-P"),
  _T("ADM-XRC-II-Lite"),
  _T("ADM-XRC-II"),
  _T("ADM-XP"),
  _T("ADM-XPL"),
  _T("ADP-WRC-II"),
  _T("ADP-DRC-II"),
  _T("ADP-XPI"),
  _T("ADM-XRC-4LS"),
  _T("ADM-XRC-4LX"),
  _T("ADM-XRC-4SX"),
  _T("ADPE-XRC-4FX"),
  _T("ADM-XRC-4FX"),
  _T("ADM-XRC-5LX"),
  _T("ADM-XRC-5T1"),
  _T("ADM-XRC-5T2"),
  _T("ADCP-XRC-4LX"),
  _T("ADM-AMC-5A2"),
  _T("ADM-XRC-5TZ"),
  _T("ADC-BBP"),
  _T("ADM-XRC-5T2-ADV"),
  _T("ADM-XRC-5T-DA1"),
  _T("ADM-XRC-5T2-ADV6"),
  _T("ADM-XRC-5T2-ADV-CC1"),
  _T("ADM-PCIE-6S1")
};

static TCHAR packageNames[][32] = {
  _T("BG560"),
  _T("BG560"),
  _T("FG456"),
  _T("FF1152"),
  _T("FF1704"),
  _T("FF896"),
  _T("FF1517"),
  _T("FF1152"),
  _T("FF1704"),
  _T("FF1148"),
  _T("FF1148"),
  _T("FF1148"),
  _T("FF1517"),
  _T("FF1517"),
  _T("FF1153"),
  _T("FF1136"),
  _T("FF1738"),
  _T("FF1148"),
  _T("FF1738"),
  _T("FF1738"),
  _T("(TBD)"),
  _T("FF1738"),
  _T("FF1136"),
  _T("FF1738"),
  _T("FF1738"),
  _T("FGG484")
};

static TCHAR deviceNames[][32] = {
  _T("XC4085XL"),
  _T("XC40150XV"),
  _T("XC40200XV"),
  _T("XC40250XV"), 
  _T("Virtex V1000"),
  _T("Virtex V400"),
  _T("Virtex V600"),
  _T("Virtex V800"), 
  _T("Virtex V2000E"),
  _T("Virtex V1000E"),
  _T("Virtex V1600E"),
  _T("Virtex V3200E"),
  _T("Virtex V812E"),
  _T("Virtex V405E"),
  _T("Unknown14"),
  _T("Unknown15"),
  _T("Unknown16"),
  _T("Unknown17"),
  _T("Unknown18"),
  _T("Unknown19"),
  _T("Unknown20"),
  _T("Unknown21"),
  _T("Unknown22"),
  _T("Unknown23"),
  _T("Unknown24"),
  _T("Unknown25"),
  _T("Unknown26"),
  _T("Unknown27"),
  _T("Unknown28"),
  _T("Unknown29"),
  _T("Unknown30"),
  _T("Unknown31"),
  _T("Virtex-II 2V1000"),
  _T("Virtex-II 2V1500"),
  _T("Virtex-II 2V2000"),
  _T("Virtex-II 2V3000"),
  _T("Virtex-II 2V4000"),
  _T("Virtex-II 2V6000"),
  _T("Virtex-II 2V8000"),
  _T("Virtex-II 2V10000"),
  _T("Unknown40"),
  _T("Unknown41"),
  _T("Unknown42"),
  _T("Unknown43"),
  _T("Unknown44"),
  _T("Unknown45"),
  _T("Unknown46"),
  _T("Unknown47"),
  _T("Unknown48"),
  _T("Unknown49"),
  _T("Unknown50"),
  _T("Unknown51"),
  _T("Unknown52"),
  _T("Unknown53"),
  _T("Unknown54"),
  _T("Unknown55"),
  _T("Unknown56"),
  _T("Unknown57"),
  _T("Unknown58"),
  _T("Unknown59"),
  _T("Unknown60"),
  _T("Unknown61"),
  _T("Unknown62"),
  _T("Unknown63"),
  _T("Virtex-II Pro 2VP2"),
  _T("Virtex-II Pro 2VP4"),
  _T("Virtex-II Pro 2VP7"),
  _T("Virtex-II Pro 2VP20"),
  _T("Virtex-II Pro 2VP30"),
  _T("Virtex-II Pro 2VP40"),
  _T("Virtex-II Pro 2VP50"),
  _T("Virtex-II Pro 2VP100"),
  _T("Virtex-II Pro 2VP125"),
  _T("Virtex-II Pro 2VP70"),
  _T("Unknown74"),
  _T("Unknown75"),
  _T("Unknown76"),
  _T("Unknown77"),
  _T("Unknown78"),
  _T("Unknown79"),
  _T("Unknown80"),
  _T("Unknown81"),
  _T("Unknown82"),
  _T("Unknown83"),
  _T("Unknown84"),
  _T("Unknown85"),
  _T("Unknown86"),
  _T("Unknown87"),
  _T("Unknown88"),
  _T("Unknown89"),
  _T("Unknown90"),
  _T("Unknown91"),
  _T("Unknown92"),
  _T("Unknown93"),
  _T("Unknown94"),
  _T("Unknown95"),
  _T("Virtex-4 V4LX15"),
  _T("Virtex-4 V4LX25"),
  _T("Virtex-4 V4LX40"),
  _T("Virtex-4 V4LX60"),
  _T("Virtex-4 V4LX100"),
  _T("Virtex-4 V4LX160"),
  _T("Virtex-4 V4LX200"),
  _T("Virtex-4 V4LX80"),
  _T("Virtex-4 V4SX25"),
  _T("Virtex-4 V4SX35"),
  _T("Virtex-4 V4SX55"),
  _T("Unknown107"),
  _T("Unknown108"),
  _T("Unknown109"),
  _T("Unknown110"),
  _T("Unknown111"),
  _T("Virtex-4 V4FX12"),
  _T("Virtex-4 V4FX20"),
  _T("Virtex-4 V4FX40"),
  _T("Virtex-4 V4FX60"),
  _T("Virtex-4 V4FX100"),
  _T("Virtex-4 V4FX140"),
  _T("Unknown118"),
  _T("Unknown119"),
  _T("Unknown120"),
  _T("Unknown121"),
  _T("Unknown122"),
  _T("Unknown123"),
  _T("Unknown124"),
  _T("Unknown125"),
  _T("Unknown126"),
  _T("Unknown127"),
  _T("Virtex-5 V5LX30"),
  _T("Virtex-5 V5LX50"),
  _T("Virtex-5 V5LX85"),
  _T("Virtex-5 V5LX110"),
  _T("Virtex-5 V5LX220"),
  _T("Virtex-5 V5LX330"),
  _T("Virtex-5 V5LX155"),
  _T("Unknown135"),
  _T("Virtex-5 V5LX30T"),
  _T("Virtex-5 V5LX50T"),
  _T("Virtex-5 V5LX85T"),
  _T("Virtex-5 V5LX110T"),
  _T("Virtex-5 V5LX330T"),
  _T("Virtex-5 V5LX220T"),
  _T("Virtex-5 V5LX155T"),
  _T("Unknown143"),
  _T("Virtex-5 V5SX35T"),
  _T("Virtex-5 V5SX50T"),
  _T("Virtex-5 V5SX95T"),
  _T("Virtex-5 V5SX240T"),
  _T("Unknown148"),
  _T("Unknown149"),
  _T("Unknown150"),
  _T("Unknown151"),
  _T("Virtex-5 V5FX100T"),
  _T("Virtex-5 V5FX130T"),
  _T("Virtex-5 V5FX200T"),
  _T("Virtex-5 V5FX30T"),
  _T("Virtex-5 V5FX70T"),
  _T("Unknown157"),
  _T("Unknown158"),
  _T("Unknown159"),
  _T("Unknown160"),
  _T("Unknown161"),
  _T("Unknown162"),
  _T("Unknown163"),
  _T("Unknown164"),
  _T("Unknown165"),
  _T("Unknown166"),
  _T("Unknown167"),
  _T("Unknown168"),
  _T("Unknown169"),
  _T("Unknown170"),
  _T("Unknown171"),
  _T("Unknown172"),
  _T("Unknown173"),
  _T("Unknown174"),
  _T("Unknown175"),
  _T("Unknown176"),
  _T("Unknown177"),
  _T("Unknown178"),
  _T("Unknown179"),
  _T("Unknown180"),
  _T("Unknown181"),
  _T("Unknown182"),
  _T("Unknown183"),
  _T("Unknown184"),
  _T("Unknown185"),
  _T("Unknown186"),
  _T("Unknown187"),
  _T("Unknown188"),
  _T("Unknown189"),
  _T("Unknown190"),
  _T("Unknown191"),
  _T("Spartan-6 LX4"),
  _T("Spartan-6 LX9"),
  _T("Spartan-6 LX16"),
  _T("Spartan-6 LX25"),
  _T("Spartan-6 LX45"),
  _T("Spartan-6 LX75"),
  _T("Spartan-6 LX100"),
  _T("Spartan-6 LX150"),
  _T("Spartan-6 LX25T"),
  _T("Spartan-6 LX45T"),
  _T("Spartan-6 LX75T"),
  _T("Spartan-6 LX100T"),
  _T("Spartan-6 LX150T")
};

static const TCHAR*
getSpaceDesc(
  int index,
  ADMXRC2_BOARD_TYPE boardtype)
{
  if (index == 0) {
    return _T("FPGA");
  } else {
    switch (boardtype) {
    case ADMXRC2_BOARD_ADMXRC:
    case ADMXRC2_BOARD_ADMXRC_P:
    case ADMXRC2_BOARD_ADMXRC2_LITE:
    case ADMXRC2_BOARD_ADMXRC2:
    case ADMXRC2_BOARD_ADMXPL:
    case ADMXRC2_BOARD_ADMXP:
    case ADMXRC2_BOARD_ADPWRC2:
    case ADMXRC2_BOARD_ADPDRC2:
    case ADMXRC2_BOARD_ADPXPI:
    case ADMXRC2_BOARD_ADMXRC4LS:
    case ADMXRC2_BOARD_ADMXRC4LX:
    case ADMXRC2_BOARD_ADMXRC4SX:
    case ADMXRC2_BOARD_ADMXRC4FX:
    case ADMXRC2_BOARD_ADPEXRC4FX:
    case ADMXRC2_BOARD_ADMXRC5LX:
    case ADMXRC2_BOARD_ADMXRC5T1:
    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADCPXRC4LX:
    case ADMXRC2_BOARD_ADMAMC5A2:
    case ADMXRC2_BOARD_ADMXRC5TZ:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5TDA1:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
    case ADMXRC2_BOARD_ADMPCIE6S1:
      if (index == 1) {
        return _T("control");
      } else {
        return _T("unknown!");
      }

    default:
      return _T("unknown!");
    }
  }
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  int ret = 0;
  ADMXRC2_HANDLE hCard = ADMXRC2_HANDLE_INVALID_VALUE;
  ADMXRC2_STATUS status;
  ADMXRC2_VERSION_INFO versionInfo;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_SPACE_INFO* pSpaces;
  TCHAR deviceName[256];
  TCHAR modelName[256];
  TCHAR packageName[256];
  Option options[2];
  Arguments arguments;
  unsigned int i;

  options[0].key = _T("card");
  options[0].type = Option_uint;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("ID of card to use");
  options[1].key = _T("index");
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("index of card to use");
  arguments.nOption = 2;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  parse_command_line(_T("INFO utility"), _T("info"), argc, argv, &arguments);

  if (options[0].specified && options[1].specified) {
    _tprintf(_T("Do not specify both -card and -index\n"));
    ret = -1;
    goto done;
  }

  if (!options[1].specified) {
    ADMXRC2_CARDID cardID = options[0].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with ID %lu: %s\n"), (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[1].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  status = ADMXRC2_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  status = ADMXRC2_GetVersionInfo(hCard, &versionInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("Failed to get version info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  if (cardInfo.BoardType < 0 || cardInfo.BoardType >= ARRAY_LENGTH(modelNames)) {
    _stprintf_s(modelName, ARRAY_LENGTH(modelName), _T("unknown (0x%lX)"), (unsigned long)cardInfo.BoardType);
  } else {
    _tcscpy_s(modelName, ARRAY_LENGTH(modelName), modelNames[cardInfo.BoardType]);
  }
  if (cardInfo.BoardType < 0 || cardInfo.BoardType >= ARRAY_LENGTH(packageNames)) {
    _tcscpy_s(packageName, ARRAY_LENGTH(packageName), _T("unknown package"));
  } else {
    _tcscpy_s(packageName, ARRAY_LENGTH(packageName), packageNames[cardInfo.BoardType]);
  }
  if (cardInfo.FPGAType < 0 || cardInfo.FPGAType >= ADMXRC2_FPGA_UNKNOWN) {
    _stprintf_s(deviceName, ARRAY_LENGTH(deviceName), _T("unknown (0x%lX)"), (unsigned long)cardInfo.FPGAType);
  } else {
    _tcscpy_s(deviceName, ARRAY_LENGTH(deviceName), deviceNames[cardInfo.FPGAType]);
  }

  _tprintf(_T("API ver                      %lu.%lu\n"),
    (unsigned long) versionInfo.APIMajor,
    (unsigned long) versionInfo.APIMinor);
  _tprintf(_T("Driver ver                   %lu.%lu\n"),
    (unsigned long) versionInfo.DriverMajor,
    (unsigned long) versionInfo.DriverMinor);
  _tprintf(_T("\n"));

  _tprintf(_T("Board Type                   %s\n"), modelName);
  _tprintf(_T("Card ID                      %lu (0x%04lX)\n"),
    (unsigned long) cardInfo.CardID,
    (unsigned long) cardInfo.CardID);
  _tprintf(_T("Serial Number                %lu (0x%08lX)\n"),
    (unsigned long) cardInfo.SerialNum,
    (unsigned long) cardInfo.SerialNum);
  _tprintf(_T("Board/Logic Rev              %u.%u/%u.%u\n"),
    cardInfo.BoardRevision >> 4, cardInfo.BoardRevision & 0xF,
    cardInfo.LogicRevision >> 4, cardInfo.LogicRevision & 0xF);
  _tprintf(_T("FPGA                         %s [%s]\n"), deviceName, packageName);
  _tprintf(_T("\n"));

  _tprintf(_T("Number of clock generators   %lu\n"), (unsigned long)cardInfo.NumClock);
  _tprintf(_T("Number of DMA channels       %lu\n"), (unsigned long)cardInfo.NumDMAChan);
  _tprintf(_T("Number of spaces             %lu\n"), (unsigned long)cardInfo.NumSpace);
  _tprintf(_T("\n"));

  pSpaces = (ADMXRC2_SPACE_INFO*) malloc(cardInfo.NumSpace * sizeof(ADMXRC2_SPACE_INFO));
  for (i = 0; i < cardInfo.NumSpace; i++) {
    unsigned long lsize, vsize;

    status = ADMXRC2_GetSpaceInfo(hCard, i, &pSpaces[i]);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Space %d: unable to get information\n"), i);
    } else {
      _tprintf(_T("Space %d (%s):\n"), i, getSpaceDesc(i, cardInfo.BoardType));
      _tprintf(_T("Physical base                0x%08lX\n"), (unsigned long)pSpaces[i].PhysicalBase);
      lsize = pSpaces[i].LocalSize;
      if (lsize) {
        _tprintf(_T("Local range                  0x%08lX - 0x%08lX\n"),
          (unsigned long)pSpaces[i].LocalBase,
          (unsigned long)(pSpaces[i].LocalBase + pSpaces[i].LocalSize - 1));
      }
      vsize = pSpaces[i].VirtualSize;
      if (vsize) {
        if (sizeof(void*) == 4) {
          uint32_t tmpBase = (uint32_t)(uintptr_t)pSpaces[i].VirtualBase;
          uint32_t tmpLimit = (uint32_t)(uintptr_t)((uint8_t*)pSpaces[i].VirtualBase + vsize - 1);
          _tprintf(_T("Virtual range                0x%08lX - 0x%08lX\n"),
            (unsigned long)tmpBase, (unsigned long)tmpLimit);
        } else if (sizeof(uintptr_t) == 8) {
          uint64_t tmpBase = (uint64_t)(uintptr_t)pSpaces[i].VirtualBase;
          uint64_t tmpLimit = (uint64_t)(uintptr_t)((uint8_t*)pSpaces[i].VirtualBase + vsize - 1);
          _tprintf(_T("Virtual range                0x%08lX_%08lX - 0x%08lX_%08lX\n"),
            (unsigned long)((tmpBase >> 32) & 0xFFFFFFFFU), (unsigned long)((tmpBase >> 0) & 0xFFFFFFFFU),
            (unsigned long)((tmpLimit >> 32) & 0xFFFFFFFFU), (unsigned long)((tmpLimit >> 0) & 0xFFFFFFFFU));
        } else {
          assert(FALSE);
        }
      }
    }
  }
  free(pSpaces);
  _tprintf(_T("\n"));

  _tprintf(_T("Number of RAM banks          %lu\n"), (unsigned long)cardInfo.NumRAMBank);
  _tprintf(_T("Bank presence bitmap         0x%08lX\n"), (unsigned long)cardInfo.RAMBanksFitted);
  for (i = 0; i < cardInfo.NumRAMBank; i++) {
    ADMXRC2_BANK_INFO bank;

    _tprintf(_T("RAM Bank %02lu                  "), (unsigned long)i);
    status = ADMXRC2_GetBankInfo(hCard, i, &bank);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("Cannot determine\n"));
    } else {
      if (!bank.Fitted) {
        _tprintf(_T("Not fitted\n"));
      } else {
        uint32_t size, byteWidth;
        TCHAR type[256];
        int ntype = 0;

        _tcscpy_s(type, ARRAY_LENGTH(type), _T(""));
        if (bank.Type & ADMXRC2_RAM_ZBTFT) {
          if (ntype) {
            _tcscat_s(type, ARRAY_LENGTH(type), _T("+"));
          }
          _tcscat_s(type, ARRAY_LENGTH(type), _T("ZBTFT"));
          ntype++;
        }
        if (bank.Type & ADMXRC2_RAM_ZBTP) {
          if (ntype) {
            _tcscat_s(type, ARRAY_LENGTH(type), _T("+"));
          }
          _tcscat_s(type, ARRAY_LENGTH(type), _T("ZBTP"));
          ntype++;
        }
        if (bank.Type & ADMXRC2_RAM_SDRAM_DDR) {
          if (ntype) {
            _tcscat_s(type, ARRAY_LENGTH(type), _T("+"));
          }
          _tcscat_s(type, ARRAY_LENGTH(type), _T("DDR SDRAM"));
          ntype++;
        }
        if (bank.Type & ADMXRC2_RAM_SRAM_DDR2) {
          if (ntype) {
            _tcscat_s(type, ARRAY_LENGTH(type), _T("+"));
          }
          _tcscat_s(type, ARRAY_LENGTH(type), _T("DDR-II SRAM"));
          ntype++;
        }
        if (bank.Type & ADMXRC2_RAM_SDRAM_SDR) {
          if (ntype) {
            _tcscat_s(type, ARRAY_LENGTH(type), _T("+"));
          }
          _tcscat_s(type, ARRAY_LENGTH(type), _T("SDR SDRAM"));
          ntype++;
        }
        if (bank.Type & ADMXRC2_RAM_SDRAM_DDR2) {
          if (ntype) {
            _tcscat_s(type, ARRAY_LENGTH(type), _T("+"));
          }
          _tcscat_s(type, ARRAY_LENGTH(type), _T("DDR-II SDRAM"));
          ntype++;
        }
        if ((bank.Width % 9) == 0) {
          /* Take care of x9 memories (with parity bits) */
          byteWidth = bank.Width / 9;
        } else {
          byteWidth = bank.Width / 8;
        }
        size = byteWidth * (bank.Size >> 10);
        _tprintf(_T("%s %lu kiWord x %lubits (%lu kiB)\n"),
          type,
          (unsigned long)(bank.Size >> 10),
          (unsigned long)bank.Width,
          (unsigned long)size);
      }
    }
  }
  _tprintf(_T("\n"));

done:
  if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(hCard);
  }

  return ret;
}
