/*
** loader.c - Utility that configures the target FPGA with a specified .bit
**            file and then exits.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XPL
**   o ADM-XP
**   o ADP-WRC-II
**   o ADP-DRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**   o ADM-PCIE-6S1
**
** (C) Copyright 2013 Alpha Data
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_WIN32)

/* Windows */
#include <windows.h>
#include <tchar.h>

#else

#include <strings.h>

/* Linux or VxWorks */
# define _T(x) x
# define _tmain main
# define _tprintf printf

#endif

#include <admxrc2.h>

#include <sdk_common.h>

int
_tmain(
  int argc,
  TCHAR *argv[])
{
	int ret = 0;
	ADMXRC2_STATUS status;
	ADMXRC2_HANDLE card = ADMXRC2_HANDLE_INVALID_VALUE;
	ADMXRC2_CARD_INFO cardInfo;
	Option options[2];
	OptionType paramType[1];
	TCHAR* paramInfo[1];
	TCHAR* paramMeaning[1];
	Arguments arguments;

	/* Set up expected options */
	options[0].key = _T("card");
	options[0].type = Option_uint;
	options[0].def.uintVal = 0;
	options[0].meaning = _T("ID of card to use");
	options[1].key = _T("index");
	options[1].type = Option_uint;
	options[1].def.uintVal = 0;
	options[1].meaning = _T("index of card to use");
	arguments.nOption = 2;
	arguments.option = options;

	paramType[0] = Option_path;
	paramInfo[0] = _T("file");
	paramMeaning[0] = _T("Bitstream filename");
	arguments.nParamType = sizeof(paramType) / sizeof(OptionType);
	arguments.minNParam = 1;
	arguments.paramType = paramType;
	arguments.paramInfo = paramInfo;
	arguments.paramMeaning = paramMeaning;

	/* Parse command line */
	parse_command_line(_T("LOADER utility"), _T("loader"), argc, argv, &arguments);

	if (options[0].specified && options[1].specified) {
		_tprintf(_T("Do not specify both -card and -index\n"));
		ret = -1;
		goto done;
	}

	if (!options[1].specified) {
		ADMXRC2_CARDID cardID = options[0].value.uintVal;

		status = ADMXRC2_OpenCard(cardID, &card);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with ID %lu: %s\n"),	(unsigned long)cardID, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
		unsigned int index = options[1].value.uintVal;

		status = ADMXRC2_OpenCardByIndex(index, &card);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with index %lu: %s\n"), (unsigned long)index, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	}

	status = ADMXRC2_GetCardInfo(card, &cardInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	switch (cardInfo.BoardType) {
	case ADMXRC2_BOARD_ADMXRC:
	case ADMXRC2_BOARD_ADMXRC_P:
	case ADMXRC2_BOARD_ADMXRC2_LITE:
	case ADMXRC2_BOARD_ADMXRC2:
	case ADMXRC2_BOARD_ADPWRC2:
	case ADMXRC2_BOARD_ADPDRC2:
	case ADMXRC2_BOARD_ADMXRC4LS:
	case ADMXRC2_BOARD_ADMXRC4LX:
	case ADMXRC2_BOARD_ADMXRC4SX:
  case ADMXRC2_BOARD_ADMXRC5TZ:
	case ADMXRC2_BOARD_ADMXRC5TDA1:
	case ADMXRC2_BOARD_ADMXPL:
	case ADMXRC2_BOARD_ADMXP:
	case ADMXRC2_BOARD_ADPXPI:
	case ADMXRC2_BOARD_ADMXRC4FX:
	case ADMXRC2_BOARD_ADPEXRC4FX:
	case ADMXRC2_BOARD_ADMXRC5LX:
	case ADMXRC2_BOARD_ADMXRC5T1:
  case ADMXRC2_BOARD_ADMXRC5T2:
  case ADMXRC2_BOARD_ADMXRC5T2ADV:
  case ADMXRC2_BOARD_ADMXRC5T2ADV6:
  case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
	case ADMXRC2_BOARD_ADMPCIE6S1:
    break;

	default:
		_tprintf(_T("*** This example must be run on one of the following cards:\n\n"));
		_tprintf(_T("\tADM-XRC\n"));
		_tprintf(_T("\tADM-XRC-P\n"));
		_tprintf(_T("\tADM-XRC-II-Lite\n"));
		_tprintf(_T("\tADM-XRC-II\n"));
		_tprintf(_T("\tADM-XPL\n"));
		_tprintf(_T("\tADM-XP\n"));
		_tprintf(_T("\tADP-WRC-II\n"));
		_tprintf(_T("\tADP-DRC-II\n"));
		_tprintf(_T("\tADP-XPI\n"));
		_tprintf(_T("\tADM-XRC-4LX\n"));
		_tprintf(_T("\tADM-XRC-4SX\n"));
		_tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
		_tprintf(_T("\tADPE-XRC-4FX\n"));
		_tprintf(_T("\tADM-XRC-5LX\n"));
		_tprintf(_T("\tADM-XRC-5T1\n"));
		_tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
		_tprintf(_T("\tADM-XRC-5TZ\n"));
		_tprintf(_T("\tADM-XRC-5T-DA1\n"));
		_tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
		_tprintf(_T("\tADM-PCIE-6S1\n"));
		ret = -1;
		goto done;
	}

  status = ADMXRC2_ConfigureFromFile(card, arguments.param[0].pathVal);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to load the bitstream '%s': %s\n"), arguments.param[0].pathVal, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

done:
	if (card != ADMXRC2_HANDLE_INVALID_VALUE) {
		ADMXRC2_CloseCard(card);
	}
	
	return ret;
}
