/*
** master.c - Example demonstrating direct master cycles initiated
**            by the FPGA.
**
** Example for:
**
**   o ADM-PCIE-6S1
**
** (C) Copyright 2013-2014 Alpha Data
**
**
** NOTES ON OPERATION OF THIS EXAMPLE
**
** ADM-PCIE-6S1:
** -------------
**
** - For this example to work, the ADM-XRC Driver must create a Common Buffer
**   of at least 64 (0x40) bytes. To have the driver create a Common Buffer,
**   set the DWORD registry value
**
**     HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\admxrc\Parameters\CommonBufferSize
**
**   to 64 (0x40). Larger values are permitted, but values that are too large
**   will result in the driver failing to allocate anything, because the
**   Windows kernel allocates Common Buffer space from a special pool of
**   limited size.
**
**   Alternatively, the registry file
**
**     %ADMXRC_SDK5%\apps\win32\src\master\CommonBufferSize.reg
**
**   may be imported into the test machine's registry.
**
**   After setting CommonBufferSize to the desired value, restart the driver
**   by disabling and then reenabling the ADM-PCIE-6S1 device in Windows Device
**   Manager. This causes the value of CommonBufferSize to take effect.
**
**   NOTE: Each time the ADM-XRC Driver is installed, the CommonBufferSize
**   registry value is overwritten with 0. Hence, setting CommonBufferSize must
**   be done AFTER installing the ADM-XRC Driver.
**
** - The Direct Master to PCI Express Window is at local addr 0xFF000000 to
**   0xFF7FFFFF (8 MiB).
**
** - Sets PLX DMPBAM register as needed, by instructing the FPGA to do so, in
**   order to move the destination PCI address of the Direct Master Window
**   around.
**
** - LA[31:24] are pulled up on board, and FPGA cannot drive them. So when
**   the FPGA drives the local bus, it drives LA[23:2] and must rely on
**   LA[31:24] being pulled up.
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _tmain main
# define _T(c) c
# define _tprintf printf
# define _tscanf_s(buffer, ...) scanf(buffer, __VA_ARGS__)
# define _gettchar getchar
# define _istprint isprint
# define _totlower tolower

#endif

#include <admxrc2.h>
#include <admxrc2/combuff.h>

#include <sdk_common.h>

#include <platform_endian.h>
#include <platform_time.h>

/* PLX register offsets (from PCI side) */
#define PLX_PCI_DMRR     (0x1CU)
#define PLX_PCI_DMLBAM   (0x20U)
#define PLX_PCI_DMPBAM   (0x28U)

/* PLX register offsets (from local bus side) */
#define PLX_LOCAL_DMPBAM (0xA8U)

/* Minimum common buffer size required by this example */
#define MINIMUM_BUFF_SIZE (0x40U)

static TCHAR
menu(
  void)
{
	static bool_t bWaitForNewLine = FALSE;
	int c;

	_tprintf(_T("Enter command, (i)nit buffer, (r)read, (s)how buffer, (w)rite or (q)uit: "));
	while (1) {
		c = _gettchar();
		if (c == EOF) {
			c = _T('q');
			break;
		} else {
			if (c == _T('\n')) {
				bWaitForNewLine = FALSE;
			} else {
				if (!bWaitForNewLine) {
					if (_istprint(c)) {
						bWaitForNewLine = TRUE;
						break;
					}
				}
			}
		}
	}

	return _totlower((char) c);
}

static ADMXRC2_STATUS
readBridgeRegFromHost(
  ADMXRC2_HANDLE hCard,
  unsigned int regOffset,
  uint32_t* pValue)
{
  return ADMXRC2_Read(hCard, ADMXRC2_IOWIDTH_32, ADMXRC2_IOADAPTER, regOffset, pValue, sizeof(*pValue));
}

static ADMXRC2_STATUS
writeBridgeRegFromHost(
  ADMXRC2_HANDLE hCard,
  unsigned int regOffset,
  uint32_t value)
{
  return ADMXRC2_Write(hCard, ADMXRC2_IOWIDTH_32, ADMXRC2_IOADAPTER, regOffset, &value, sizeof(value));
}

static void
initBuffer(
  const ADMXRC2_COMMONBUFFER_INFO* pCommonBufferInfo)
{
  uint32_t* pBuf = (uint32_t*)pCommonBufferInfo->VirtualBase;
  uint32_t i;

	pBuf[0] = 0xdeadbeef;
	pBuf[1] = 0xcafeface;
	for (i = 2U; i < (MINIMUM_BUFF_SIZE >> 2); i++) {
		pBuf[i] = i;
	}
}

static void
showBuffer(
  const ADMXRC2_COMMONBUFFER_INFO* pCommonBufferInfo)
{
  uint32_t* pBuf = (uint32_t*)pCommonBufferInfo->VirtualBase;
  uint32_t i, j;

	_tprintf(_T("  --> contents of user buffer:\n"));
	for (i = 0; i < (MINIMUM_BUFF_SIZE >> 2);) {
		_tprintf(_T("      offset 0x%02lX: "), (unsigned long)(i << 2));
		for (j = 0; j < 4U; j++, i++) {
			_tprintf(_T("%08lX "), (unsigned long)pBuf[i]);
		}
		_tprintf(_T("\n"));
	}
	_tprintf(_T("\n"));
}

static void
setDMPBAM(
  volatile uint32_t* pFpgaSpace,
  uint32_t dmWindowSize,
  uint32_t pciAddr)
{
  uint32_t dmWindowMask = dmWindowSize - 1U;
  uint32_t page, val32;

	page = pciAddr & ~dmWindowMask;

	/* Select Write-to-PLX-Regs operation. */
	pFpgaSpace[2] = cpu_to_le32(0x3);
	pFpgaSpace[2];

	/* Supply data to write to DMPBAM. */
  val32 = page | 0x1U;
	pFpgaSpace[1] = cpu_to_le32(val32);
	pFpgaSpace[1];

	/* Write local address to make the FPGA perform the write. */
  val32 = PLX_LOCAL_DMPBAM;
	pFpgaSpace[0] = cpu_to_le32(val32);
	pFpgaSpace[0];

  /* Wait for FPGA's local bus cycle to be completed */
  do {
    val32 = pFpgaSpace[4];
    val32 = le32_to_cpu(val32);
  } while (!(val32 & 0x1U));
}

static void
doRead(
  ADMXRC2_HANDLE hCard,
  volatile uint32_t* pFpgaSpace,
  const ADMXRC2_COMMONBUFFER_INFO* pCommonBufferInfo,
  uint32_t dmWindowSize,
	uint32_t index,
	uint32_t* pData32)
{
  ADMXRC2_STATUS status;
  uint32_t dmWindowMask = dmWindowSize - 1U;
	uint32_t pciAddr, offset, val32;

	offset = index * sizeof(uint32_t);
#if defined(_WIN32)
  pciAddr = pCommonBufferInfo->PhysicalBase.LowPart + offset;
#else
  pciAddr = (uint32_t)pCommonBufferInfo->PhysicalBase + offset;
#endif

	/* Set up paging bits of DMPBAM in PLX chip */
	setDMPBAM(pFpgaSpace, dmWindowSize, pciAddr);

	/* Select Read-to-PCI operation */
	pFpgaSpace[2] = cpu_to_le32(0x0);
	pFpgaSpace[2];

  /* Make sure the changes to the buffer will be visible to bus masters */
  /* Not really necessary on architectures like x86/AMD64 which generally have cache-coherent DMA */
  status = ADMXRC2_SyncCommonBuffer(hCard, 0, 0, MINIMUM_BUFF_SIZE, ADMXRC2_SYNC_CPUTOFPGA);
  if (ADMXRC2_SUCCESS != status) {
    _tprintf(_T("+++ Warning: syncing common buffer (CPU to FPGA) failed: %s\n"), ADMXRC2_GetStatusString(status));
  }

	/* Write local address to make the FPGA perform the read. */
	val32 = pciAddr & dmWindowMask;
	pFpgaSpace[0] = cpu_to_le32(val32);
	pFpgaSpace[0];

  /* Wait for FPGA's local bus cycle to be completed */
  do {
    val32 = pFpgaSpace[4];
    val32 = le32_to_cpu(val32);
  } while (!(val32 & 0x1U));

	/* Read back the data read by the FPGA */
  val32 = pFpgaSpace[3];
	*pData32 = le32_to_cpu(val32);
}

static void
doWrite(
  ADMXRC2_HANDLE hCard,
  volatile uint32_t* pFpgaSpace,
  const ADMXRC2_COMMONBUFFER_INFO* pCommonBufferInfo,
  uint32_t dmWindowSize,
	uint32_t index,
	uint32_t data32)
{
  ADMXRC2_STATUS status;
  uint32_t dmWindowMask = dmWindowSize - 1U;
	uint32_t pciAddr, offset, val32;

	offset = index * sizeof(uint32_t);
#if defined(_WIN32)
  pciAddr = pCommonBufferInfo->PhysicalBase.LowPart + offset;
#else
  pciAddr = (uint32_t)pCommonBufferInfo->PhysicalBase + offset;
#endif

	/* Set up paging bits of DMPBAM in PLX chip */
	setDMPBAM(pFpgaSpace, dmWindowSize, pciAddr);

	/* Select Write-to-PCI operation */
	pFpgaSpace[2] = cpu_to_le32(0x1);
	pFpgaSpace[2];

	/* Supply data to write */
	pFpgaSpace[1] = cpu_to_le32(data32);
	pFpgaSpace[1];

	/* Write local address to make the FPGA perform the write. */
	val32 = pciAddr & dmWindowMask;
	pFpgaSpace[0] = cpu_to_le32(val32);
	pFpgaSpace[0];

  /* Wait for FPGA's local bus cycle to be completed */
  do {
    val32 = pFpgaSpace[4];
    val32 = le32_to_cpu(val32);
  } while (!(val32 & 0x1U));

  /* Make sure the changes to the buffer by bus masters will be visible to the CPU */
  /* Not really necessary on architectures like x86/AMD64 which generally have cache-coherent DMA */
  status = ADMXRC2_SyncCommonBuffer(hCard, 0, 0, MINIMUM_BUFF_SIZE, ADMXRC2_SYNC_FPGATOCPU);
  if (ADMXRC2_SUCCESS != status) {
    _tprintf(_T("+++ Warning: syncing common buffer (FPGA to CPU) failed: %s\n"), ADMXRC2_GetStatusString(status));
  }
}

int
main(int argc, TCHAR *argv[])
{
	const double clockFreq = 33.0e6;
	int ret = 0;
  ADMXRC2_HANDLE hCard = ADMXRC2_HANDLE_INVALID_VALUE;
	ADMXRC2_STATUS status;
	ADMXRC2_CARD_INFO cardInfo;
	ADMXRC2_SPACE_INFO spInfo;
  ADMXRC2_COMMONBUFFER_INFO commonBuffInfo;
	bool_t bQuit;
	const TCHAR* pFilename;
  volatile uint32_t* pFpgaSpace;
	TCHAR choice;
	Arguments arguments;
	Option options[2];
  uint32_t directMasterWindowSize, directMasterWindowLocalBase;
  uint32_t dmrrValue, dmlbamValue;

	/* Set up expected options */
	options[0].key = _T("card");
	options[0].type = Option_uint;
	options[0].def.uintVal = 0;
	options[0].meaning = _T("ID of card to use");
	options[1].key = _T("index");
	options[1].type = Option_uint;
	options[1].def.uintVal = 0;
	options[1].meaning = _T("index of card to use");
	arguments.nOption = 2;
	arguments.option = options;
	arguments.nParamType = 0;
	arguments.minNParam = 0;

	/* Parse command line */
	parse_command_line(
		_T("MASTER example"),
    _T("master"),
		argc,
		argv,
		&arguments);

	if (options[0].specified && options[1].specified) {
		_tprintf(_T("Do not specify both -card and -index\n"));
		ret = -1;
		goto done;
	}

	if (!options[1].specified) {
		ADMXRC2_CARDID cardID = options[0].value.uintVal;

		status = ADMXRC2_OpenCard(cardID, &hCard);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with ID %lu: %s\n"), (unsigned long)cardID, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	} else {
		unsigned int index = options[1].value.uintVal;

		status = ADMXRC2_OpenCardByIndex(index, &hCard);
		if (status != ADMXRC2_SUCCESS) {
			_tprintf(_T("Failed to open card with index %u: %s\n"), index, ADMXRC2_GetStatusString(status));
			ret = -1;
			goto done;
		}
	}

	status = ADMXRC2_GetCardInfo(hCard, &cardInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}

	switch (cardInfo.BoardType) {
	case ADMXRC2_BOARD_ADMPCIE6S1:
    /* Direct Master Local Bus to PCIe Window parameters */
    directMasterWindowSize = 0x800000U;        /* Local to PCI Express window size is 8 MiB */
    directMasterWindowLocalBase = 0xFF000000U; /* Local to PCI Express window local base is 0xFF000000 */
    break;

	default:
		_tprintf(_T("This example must be run on one of the following cards:\n\n"));
		_tprintf(_T("\to ADM-PCIE-6S1\n"));
		ret = -1;
		goto done;
	}

	status = ADMXRC2_GetSpaceInfo(hCard, 0, &spInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to get space 0 info: %s\n"), ADMXRC2_GetStatusString(status));
		return -1;
	}
	pFpgaSpace = (volatile uint32_t*)spInfo.VirtualBase;
	if (pFpgaSpace == NULL) {
		_tprintf(_T("*** Failed to retrieve pointer to FPGA space\n"));
		return -1;
	}

  status = ADMXRC2_GetCommonBufferInfo(hCard, 0, &commonBuffInfo);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to get common buffer info: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
  if (commonBuffInfo.Size < MINIMUM_BUFF_SIZE) {
		_tprintf(_T("Common buffer too small, require >= %lu(0x%lX) bytes; actual size is %lu(0x%lX) bytes(s)\n"),
      (unsigned long)MINIMUM_BUFF_SIZE, (unsigned long)MINIMUM_BUFF_SIZE, commonBuffInfo.Size, commonBuffInfo.Size);
		ret = -1;
		goto done;
  }
#if defined(_WIN32)
  if (commonBuffInfo.PhysicalBase.HighPart != 0U) {
#else
  if ((commonBuffInfo.PhysicalBase >> 32) != 0U) {
#endif
		_tprintf(_T("Common buffer physical address must be in low 4 GiB; actual PhysicalBase=0x%llX\n"),
#if defined(_WIN32)
      (unsigned long long)commonBuffInfo.PhysicalBase.QuadPart);
#else
      (unsigned long long)commonBuffInfo.PhysicalBase);
#endif
		ret = -1;
		goto done;
  }
  _tprintf(_T("Common buffer: PhysicalBase=0x%llX VirtualBase=%p Size=%lu(0x%lX) bytes\n"),
#if defined(_WIN32)
    (unsigned long long)commonBuffInfo.PhysicalBase.QuadPart,
#else
    (unsigned long long)commonBuffInfo.PhysicalBase,
#endif
    commonBuffInfo.VirtualBase,
    commonBuffInfo.Size,
    commonBuffInfo.Size);

  if (cardInfo.NumClock > 0) {
	  /* Set local bus clock to nominal frequency */
	  status = ADMXRC2_SetClockRate(hCard, ADMXRC2_CLOCK_LCLK, clockFreq, NULL);
	  if (status != ADMXRC2_SUCCESS) {
		  _tprintf(_T("Failed to set LCLK to %.1fMHz: %s\n"), clockFreq / 1.0e6f, ADMXRC2_GetStatusString(status));
		  ret = -1;
		  goto done;
	  }
  }

	pFilename = get_fpga_filename(&cardInfo, _T("master"), &arguments);
	status = ADMXRC2_ConfigureFromFile(hCard, pFilename);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to load the bitstream '%s': %s\n"), pFilename, ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
  sleepMs(10); /* Allow time for DLLs/DCMs/PLLs to achieve lock */

	/* Initialise the user buffer to some recognisable values */
	initBuffer(&commonBuffInfo);

  /* Initialise Direct Master Local Bus to PCIe window registers in PCI9080/9056/9656 */
  status = writeBridgeRegFromHost(hCard, PLX_PCI_DMRR, ~(directMasterWindowSize - 1U));
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to set DMRR register: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
  status = writeBridgeRegFromHost(hCard, PLX_PCI_DMLBAM, directMasterWindowLocalBase);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to set DMLBAM register: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
  status = readBridgeRegFromHost(hCard, PLX_PCI_DMRR, &dmrrValue);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to read DMRR register: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
  status = readBridgeRegFromHost(hCard, PLX_PCI_DMLBAM, &dmlbamValue);
	if (status != ADMXRC2_SUCCESS) {
		_tprintf(_T("Failed to read DMLBAM register: %s\n"), ADMXRC2_GetStatusString(status));
		ret = -1;
		goto done;
	}
  _tprintf(_T("DMRR=0x%08lX DMLBAM=0x%08lX\n"), (unsigned long)dmrrValue, (unsigned long)dmlbamValue);

	_tprintf(_T("Direct Master demonstration\n\n"));
  _tprintf(_T("DM local bus window size is 0x%lu(0x%lX) bytes\n"),
    (unsigned long)directMasterWindowSize, (unsigned long)directMasterWindowSize);

	showBuffer(&commonBuffInfo);

	bQuit = FALSE;
	while (!bQuit) {
		choice = menu();
		switch (choice) {
		case 'i':
			initBuffer(&commonBuffInfo);
			_tprintf(_T("  --> reinitialised buffer.\n\n"));
			showBuffer(&commonBuffInfo);
			break;

		case 'r':
			{
				unsigned long idx;
        uint32_t val32;

				_tprintf(_T("Enter buffer index (hex): "));
				_tscanf_s(_T("%lx"), &idx);
        if (idx >= (MINIMUM_BUFF_SIZE >> 2)) {
					_tprintf(_T("*** invalid index: 0x%lX\n"), idx);
					break;
				}
				doRead(hCard, pFpgaSpace, &commonBuffInfo, directMasterWindowSize, (uint32_t)idx, &val32);
				_tprintf(_T("  --> FPGA read 0x%08lX at index 0x%lX\n\n"), (unsigned long)val32, idx);
				break;
			}

		case 's':
			showBuffer(&commonBuffInfo);
			break;

		case 'q':
			bQuit = TRUE;
			break;

		case 'w':
			{
				unsigned long idx, val;

				_tprintf(_T("Enter buffer index (hex): "));
				_tscanf_s(_T("%lx"), &idx);
				if (idx >= (MINIMUM_BUFF_SIZE >> 2)) {
					_tprintf(_T("*** invalid index: 0x%lX\n"), idx);
					break;
				}
				_tprintf(_T("Enter value to write (hex): "));
				_tscanf_s(_T("%lx"), &val);
				doWrite(hCard, pFpgaSpace, &commonBuffInfo, directMasterWindowSize, (uint32_t)idx, (uint32_t)val);
				_tprintf(_T("  --> FPGA wrote 0x%08lX to index 0x%lX\n\n"), val, idx);
				break;
			}

		default:
			_tprintf(_T("Invalid option.\n"));
			break;
		}
	}

done:
	if (hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
		ADMXRC2_CloseCard(hCard);
	}

	return ret;
}
