/*
** memory.c - Example demonstrating host access to memory banks
**            on an ADM-X* series card.
**
** Example for:
**
**   o ADM-XRC
**   o ADM-XRC-P
**   o ADM-XRC-II-Lite
**   o ADM-XRC-II
**   o ADM-XP
**   o ADM-XPL
**   o ADP-DRC-II
**   o ADP-WRC-II
**   o ADP-XPI
**   o ADM-XRC-4LX
**   o ADM-XRC-4SX
**   o ADM-XRC-4FX / ADM-XMC-4FX
**   o ADPE-XRC-4FX
**   o ADM-XRC-5LX
**   o ADM-XRC-5T1
**   o ADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6
**   o ADM-XRC-5TZ
**   o ADM-XRC-5T-DA1
**   o ADM-XRC-5T2-ADV-CC1
**
** (C) Copyright 2005-2014 Alpha Data
*/

#define ZAP (0)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _tmain main
# define _T(c) c
# define _tprintf printf
# define _gettchar getchar
# define _stprintf_s snprintf
# define _tcscpy_s(pDst, n, pSrc) do { strncpy(pDst, pSrc, n); (pDst)[(n) - 1] = _T('\0'); } while (0)
# define _tcscat_s(pDst, n, pSrc) do { strncat(pDst, pSrc, n); (pDst)[(n) - 1] = _T('\0'); } while (0)
# define _tcslen strlen
# define _totlower tolower

#endif

#include <admxrc2.h>

#include <sdk_common.h>
#include <platform_endian.h>
#include <platform_time.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define minimumOf2Macro(a, b) ((a) < (b) ? (a) : (b))

#if !defined(_WIN32)
typedef void* HANDLE;
#endif

/*
** Register indices and bit definitions
*/

#define BANK_REG                        (0x0)               /* Bank select register */
#define BANK_REG_BANKMASK               (0xfU << 0)

#define PAGE_REG                        (0x1)               /* Page register */
#define PAGE_REG_PAGEMASK               (0x7ffU << 0)

#define MEMCTL_REG                      (0x2)               /* Memory control register */
#define MEMCTL_REG_MEMRST               (0x1U << 0)

#define STATUS_REG                      (0x4)               /* Status register */
#define STATUS_REG_LCLKLOCKED           (0x1U << 0)
#define STATUS_REG_LCLKSTICKY           (0x1U << 1)
#define STATUS_REG_SHIFT_LOCKED         (8)
#define STATUS_REG_SHIFT_STICKY         (16)

#define MEMSTAT_REG                     (0x6)               /* Memory status register */
#define MEMSTAT_REG_SHIFT_TRAINED       (0)

#define MODEx_REG(n)                    (0x10 + (n))        /* Bank 'n' mode register */

#define MODEx_REG_DDR2SSRAM_BLMASK      (0x3U << 0)         /* DDR2 SSRAM burst length mask */
#define MODEx_REG_DDR2SSRAM_BL2         (0x0U << 0)         /* DDR2 SSRAM burst length 2 */
#define MODEx_REG_DDR2SSRAM_BL4         (0x1U << 0)         /* DDR2 SSRAM burst length 4 */
#define MODEx_REG_DDR2SSRAM_DLLON       (0x0U << 2)         /* DDR2 SSRAM DLL enable */
#define MODEx_REG_DDR2SSRAM_DLLOFF      (0x1U << 2)         /* DDR2 SSRAM DLL disable */

#define MODEx_REG_ZBTSSRAM_FLOWTHRU     (0x0U << 0)         /* ZBT SSRAM pipelined mode disable */
#define MODEx_REG_ZBTSSRAM_PIPELINE     (0x1U << 0)         /* ZBT SSRAM pipelined mode enable */

#define MODEx_REG_DDRSDRAM_UNBUFFERED   (0x0U << 0)
#define MODEx_REG_DDRSDRAM_REGISTERED   (0x1U << 0)
#define MODEx_REG_DDRSDRAM_X8X16        (0x0U << 1)
#define MODEx_REG_DDRSDRAM_X4           (0x1U << 1)
#define MODEx_REG_DDRSDRAM_ROWBITS_12   (0x0U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_13   (0x1U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_14   (0x2U << 2)
#define MODEx_REG_DDRSDRAM_ROWBITS_15   (0x3U << 2)
#define MODEx_REG_DDRSDRAM_COLBITS_M4   (0x0U << 4)         /* col bits = row bits - 4 */
#define MODEx_REG_DDRSDRAM_COLBITS_M3   (0x1U << 4)         /* col bits = row bits - 3 */
#define MODEx_REG_DDRSDRAM_COLBITS_M2   (0x2U << 4)         /* col bits = row bits - 2 */
#define MODEx_REG_DDRSDRAM_COLBITS_M1   (0x3U << 4)         /* col bits = row bits - 1 */
#define MODEx_REG_DDRSDRAM_BANKBITS_0   (0x0U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_1   (0x1U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_2   (0x2U << 6)
#define MODEx_REG_DDRSDRAM_BANKBITS_3   (0x3U << 6)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_1  (0x0U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_2  (0x1U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_4  (0x2U << 8)
#define MODEx_REG_DDRSDRAM_PHYSBANKS_8  (0x3U << 8)

#define MODEx_REG_DDR2SDRAM_TRFC_1Gb    (0x0U << 10)        /* Refresh cycle time for 1Gb DDR-II SDRAM part */
#define MODEx_REG_DDR2SDRAM_TRFC_256Mb  (0x1U << 10)        /* Refresh cycle time for 256Mb DDR-II SDRAM part */
#define MODEx_REG_DDR2SDRAM_TRFC_512Mb  (0x2U << 10)        /* Refresh cycle time for 512Mb DDR-II SDRAM part */
#define MODEx_REG_DDR2SDRAM_TRFC_2Gb    (0x3U << 10)        /* Refresh cycle time for 2Gb DDR-II SDRAM part */

#define SPD_REG                         (0x5)
#define SPD_REG_DO                      (0x1U << 31)
#define SPD_REG_BUSY                    (0x1U << 31)
#define SPD_REG_SHIFT_ADDR              (0)
#define SPD_REG_SHIFT_WRDATA            (8)
#define SPD_REG_SHIFT_RDDATA            (16)
#define SPD_REG_ERR                     (0x1U << 24)
#define SPD_REG_SHIFT_SLOT              (25)

/*
** Random number generator using Mersenne Twister
*/

#define MERS_N (351)
#define MERS_M (175)
#define MERS_R (19)
#define MERS_U (11)
#define MERS_S (7)
#define MERS_T (15)
#define MERS_L (17)
#define MERS_A (0xE4BD75F5)
#define MERS_B (0x655E5280)
#define MERS_C (0xFFD58000)

struct _MTState {
    uint32_t mt[MERS_N]; /* state vector */
    int mti; /* index into mt */
};
typedef struct _MTState MTState;

const unsigned long     memWindowAddr = 0x200000U;   /* In bytes */
const unsigned long     pageSize  = 0x200000U;   /* In bytes */
const unsigned long     pageShift = 21;

const unsigned long     maxReportedErr = 20;

/* Chunk size for generating pattern data to write to memory banks */
#define GENERATE_CHUNK (0x1U << 20)
#define GENERATE_CHUNK_RD (GENERATE_CHUNK)

/* Max. number of memory banks supported by this application */
#define MAX_BANK (16)

/* Max. logical width of a memory bank, in bytes */
#define MAX_BANK_WIDTH (64)

struct _Globals {
  uint8_t revLookup[0x100];
  admxrc2_time_t lastSpinTick;
  unsigned int spIndex;
  MTState mtState;
  volatile void* ppFpgaSpace;
  volatile uint32_t* pFpgaRegs;
  uint8_t* pMemoryWindow;
  uint8_t* pBuf5a;
  uint8_t* pBufOwnAddr;
  uint8_t* pBufRevAddr;
  uint8_t* pBufRandom;
  uint8_t* pBufChk;
  ADMXRC2_HANDLE hCard;
  HANDLE hDmaEvent;
  uint32_t dmaMode;
  uint32_t bankTestMask;
  bool_t bBankDdr[MAX_BANK];
  unsigned int bankWidth[MAX_BANK];
  uint64_t bankSize[MAX_BANK];
  bool_t bUse64Bit;
};

typedef struct _Globals Globals;

void
seed(
  MTState* state,
  uint32_t seed)
{
  /* re-seed generator */
  state->mt[0] = seed;
  for (state->mti = 1; state->mti < MERS_N; state->mti++) {
    state->mt[state->mti] = (1812433253UL * (state->mt[state->mti - 1] ^ (state->mt[state->mti - 1] >> 30)) + state->mti);
  }
}

static uint32_t
random32(
  MTState* state)
{
  uint32_t y;

  if (state->mti >= MERS_N) {
    /* generate MERS_N words at one time */
    const uint32_t LOWER_MASK = (1LU << MERS_R) - 1; /* lower MERS_R bits */
    const uint32_t UPPER_MASK = -1L  << MERS_R;      /* upper (32 - MERS_R) bits */
    static const uint32_t mag01[2] = { 0, MERS_A };

    int kk;
    for (kk = 0; kk < (MERS_N - MERS_M); kk++) {
      y = (state->mt[kk] & UPPER_MASK) | (state->mt[kk + 1] & LOWER_MASK);
      state->mt[kk] = state->mt[kk + MERS_M] ^ (y >> 1) ^ mag01[y & 1];
    }

    for (; kk < MERS_N-1; kk++) {    
      y = (state->mt[kk] & UPPER_MASK) | (state->mt[kk + 1] & LOWER_MASK);
      state->mt[kk] = state->mt[kk + (MERS_M - MERS_N)] ^ (y >> 1) ^ mag01[y & 1];}      

    y = (state->mt[MERS_N - 1] & UPPER_MASK) | (state->mt[0] & LOWER_MASK);
    state->mt[MERS_N - 1] = state->mt[MERS_M - 1] ^ (y >> 1) ^ mag01[y & 1];
    state->mti = 0;
  }
  y = state->mt[state->mti++];

  /* Tempering (may be omitted) */
  y ^=  y >> MERS_U;
  y ^= (y << MERS_S) & MERS_B;
  y ^= (y << MERS_T) & MERS_C;
  y ^=  y >> MERS_L;

  return y;
}

static ADMXRC2_STATUS
reverse(
  Globals* pGlob,
  uint32_t v)
{
  uint8_t a, b, c, d;

  a = pGlob->revLookup[(v >> 0) & 0xff];
  b = pGlob->revLookup[(v >> 8) & 0xff];
  c = pGlob->revLookup[(v >> 16) & 0xff];
  d = pGlob->revLookup[(v >> 24) & 0xff];
  return ((uint32_t)a << 24) | ((uint32_t)b << 16) | ((uint32_t)c << 8) | ((uint32_t)d << 0);
}

static void
spin(
  Globals* pGlobals,
  bool_t bWriting)
{
  const TCHAR spinner[] = _T("-\\|/");
  unsigned int spIndex;
  admxrc2_time_t now, nextTick;

  now = timeGet();
  nextTick = timeAdd(now, timeMake(0, 50000, TRUE));
  if (timeAfter(now, nextTick)) {
    spIndex = pGlobals->spIndex;
    _tprintf(_T("%c%c\010\010"), bWriting ? _T('W') : _T('R'), spinner[spIndex]);
    spIndex = (spIndex + 1) & 0x3U;
    pGlobals->spIndex = spIndex;
    pGlobals->lastSpinTick = now;
  }
}

static void wordToString(
  TCHAR* pStr,
  size_t max,
  uint8_t* buff,
  unsigned int n)
{
  unsigned int i;
  TCHAR s[8];

  _tcscpy_s(pStr, max, _T("0x"));
  for (i = n; i > 0;) {
    i--;
    _stprintf_s(s, ARRAY_LENGTH(s), _T("%02lX"), (unsigned long)buff[i]);
    _tcscat_s(pStr, max, s);
    if (i != 0) {
      _tcscat_s(pStr, max, _T("_"));
    }
  }
}

static bool_t
verifyWord(
  uint8_t* buff,
  uint8_t* exp,
  uint8_t* act,
  uint8_t* mask,
  unsigned int width)
{
  unsigned int i;

  for (i = 0; i < width; i++) {
    act[i] = *buff++ & mask[i];
  }
  for (i = 0; i < width; i++) {
    if (act[i] != exp[i]) {
      return TRUE;
    }
  }
  return FALSE;
}

static void
displayError(
  uint64_t location,
  uint8_t* exp,
  uint8_t* act,
  unsigned int width,
  uint64_t nErr)
{
  TCHAR strExp[MAX_BANK_WIDTH * 3 + 2];
  TCHAR strAct[MAX_BANK_WIDTH * 3 + 2];

  if (nErr <= maxReportedErr) {
    wordToString(strExp, ARRAY_LENGTH(strExp), exp, width);
    wordToString(strAct, ARRAY_LENGTH(strAct), act, width);
    if (nErr == 1) {
      _tprintf(_T("\n"));
    }
    _tprintf(_T("*** Loc 0x%01lX_%08lX: exp=%s\n"),
      (unsigned long)(location >> 32), (unsigned long)(uint32_t)location, strExp);
    _tprintf(_T("                      act=%s\n"), strAct);
  } else if (nErr == maxReportedErr + 1) {
    _tprintf(_T("*** [More errors]\n"));
  }
}

static void
displayMask(
  unsigned int bankWidth,
  uint8_t* mask,
  TCHAR* str,
  size_t max)
{
  TCHAR buff[3];
  unsigned int i;
  size_t n;

  assert(max > 0);
  str[0] = _T('\0');
  for (i = 0; i < bankWidth; i++) {
    _stprintf_s(buff, ARRAY_LENGTH(buff), _T("%02lX"), (unsigned long)mask[bankWidth - i - 1]);
    n = _tcslen(buff);
    _tcscat_s(str, max, buff);
    if (i != (bankWidth - 1)) {
    _tcscat_s(str, max, _T("_"));
    }
  }
}

static bool_t
readSpdByte(
  Globals* pGlob,
  unsigned int slot,
  unsigned int index,
  uint8_t* value)
{
  volatile uint32_t* regs = pGlob->pFpgaRegs;
  uint32_t val32;
  uint32_t stat32;
  bool_t bFail;

  val32 = (index & 0xffU) << SPD_REG_SHIFT_ADDR;
  val32 |= (slot & 0x7U) << SPD_REG_SHIFT_SLOT;
  regs[SPD_REG] = val32;
  val32 |= SPD_REG_DO;
  regs[SPD_REG] = val32;
  stat32 = regs[SPD_REG];
  while (1) {
    stat32 = regs[SPD_REG];
    if (!(stat32 & SPD_REG_BUSY)) {
      break;
    }
  }
  if (stat32 & SPD_REG_ERR) {
    bFail = TRUE;
  } else {
    bFail = FALSE;
  }
  *value = (uint8_t) (stat32 >> SPD_REG_SHIFT_RDDATA);
  return bFail;
}

static bool_t
readSpd(
  Globals* pGlob,
  unsigned int slot,
  uint8_t data[128])
{
  unsigned int i;
  bool_t bFail;

  for (i = 0; i < 128; i++) {
    bFail = readSpdByte(pGlob, slot, i, &data[i]);
    if (bFail || (i == 0 && data[i] == 0xffU)) {
      break;
    }
  }
  return (!bFail && data[0] != 0xffU);
}

static void
selectBank(
  Globals* pGlob,
  unsigned int bank)
{
  pGlob->pFpgaRegs[BANK_REG] = (uint32_t)bank & BANK_REG_BANKMASK;
  pGlob->pFpgaRegs[BANK_REG];
}

static ADMXRC2_STATUS
readMemory(
  Globals*        pGlob,
  const void*     buffer,
  uint64_t        ramOffset,
  size_t          length,
  int             useDMA)
{
  ADMXRC2_STATUS status;
  uint8_t* dst = (uint8_t*)buffer;
  uint8_t* src;
  unsigned long pgidx;
  unsigned long pgoffs;
  unsigned long chunk;

  while (length) {
    pgidx = (unsigned long)(ramOffset >> pageShift);
    pgoffs = (unsigned long)ramOffset & (pageSize - 1);

    chunk = pageSize - pgoffs;
    if (chunk > length) {
      chunk = (unsigned long)length; /* Cast is safe due to above check */
    }

    /* Set the page register */
    pGlob->pFpgaRegs[PAGE_REG] = pgidx & PAGE_REG_PAGEMASK;
    pGlob->pFpgaRegs[PAGE_REG]; /* make sure it's got there */

    if (useDMA) {
      status = ADMXRC2_DoDMAImmediate(
        pGlob->hCard,
        dst,
        chunk,
        memWindowAddr + pgoffs,
        ADMXRC2_LOCALTOPCI,
        ADMXRC2_DMACHAN_ANY,
        pGlob->dmaMode,
        0,
        NULL,
        pGlob->hDmaEvent);
      if (status != ADMXRC2_SUCCESS) {
        return status;
      }
    } else {
      src = pGlob->pMemoryWindow + pgoffs;
      memcpy(dst, src, chunk);
    }
    dst += chunk;
    ramOffset += chunk;
    length -= chunk;
  }

  return ADMXRC2_SUCCESS;
}

static ADMXRC2_STATUS
writeMemory(
  Globals*        pGlob,
  const void*     buffer,
  uint64_t        ramOffset,
  size_t          length,
  int             useDMA)
{
  ADMXRC2_STATUS status;
  uint8_t* dst;
  uint8_t* src = (uint8_t*)buffer;
  unsigned long pgidx;
  unsigned long pgoffs;
  unsigned long chunk;

  while (length) {
    pgidx = (unsigned long)(ramOffset >> pageShift);
    pgoffs = (unsigned long)ramOffset & (pageSize - 1);

    chunk = pageSize - pgoffs;
    if (chunk > length) {
      chunk = (unsigned long)length; /* Cast is safe due to above check */
    }

    /* Set the page register */
    pGlob->pFpgaRegs[PAGE_REG] = pgidx & PAGE_REG_PAGEMASK;
    pGlob->pFpgaRegs[PAGE_REG]; /* make sure it's got there */

    if (useDMA) {
      status = ADMXRC2_DoDMAImmediate(
        pGlob->hCard,
        src,
        chunk,
        memWindowAddr + pgoffs,
        ADMXRC2_PCITOLOCAL,
        ADMXRC2_DMACHAN_ANY,
        pGlob->dmaMode,
        0,
        NULL,
        pGlob->hDmaEvent);
      if (status != ADMXRC2_SUCCESS) {
        return status;
      }
    } else {
      dst = pGlob->pMemoryWindow + pgoffs;
      memcpy(dst, src, chunk);
    }
    src += chunk;
    ramOffset += chunk;
    length -= chunk;
  }

  return ADMXRC2_SUCCESS;
}

static void
maskExpectedWord(
  uint8_t* exp,
  uint8_t* data,
  uint8_t* mask,
  unsigned int width)
{
  unsigned int i;

  for (i = 0; i < width; i++) {
    exp[i] = data[i] & mask[i];
  }
}

static void
generate5A(
  uint8_t* buffer,
  unsigned long size,
  unsigned int width,
  int bIsDdr)
{
  unsigned long i;
  unsigned int j;
  int x = 0;
  uint8_t patt;
  assert(width >= 2);

  if (bIsDdr) {
    unsigned int halfwidth = width >> 1;

    assert((size % (width << 1)) == 0);
    for (i = 0; i < size;) {
      patt = x ? 0xAA : 0x55;
      for (j = 0; j < halfwidth; j++) {
        buffer[i++] = patt;
      }
      patt ^= 0xff;
      for (j = 0; j < halfwidth; j++) {
        buffer[i++] = patt;
      }
      x ^= 0x1;
    }
  } else {
    assert((size % (width << 2)) == 0);
    for (i = 0; i < size;) {
      patt = x ? 0xAA : 0x55;
      for (j = 0; j < width; j++) {
        buffer[i++] = patt;
      }
      patt ^= 0xff;
      for (j = 0; j < width; j++) {
        buffer[i++] = patt;
      }
      x ^= 0x1;
    }
  }
}

static void
generateOwnAddr(
  uint32_t* buffer,
  uint32_t offset,
  unsigned long size)
{
  unsigned long i;

  assert((offset & 0x3U) == 0);
  assert((size & 0x3U) == 0);

  size >>= 2;
  for (i = 0; i < size; i++) {
    *buffer++ = offset++;
  }
}

static void
generateRevAddr(
  Globals* pGlob,
  uint32_t* buffer,
  unsigned long offset,
  unsigned long size)
{
  unsigned long i;

  assert((offset & 0x3U) == 0);
  assert((size & 0x3U) == 0);

  size >>= 2;
  for (i = 0; i < size; i++) {
    *buffer++ = reverse(pGlob, offset);
    offset++;
  }
}

static void
generateRandom(
  MTState* mtState,
  uint32_t* buffer,
  unsigned long size)
{
  unsigned long i;

  assert((size & 0x3U) == 0);

  size >>= 2;
  for (i = 0; i < size; i++) {
    *buffer++ = random32(mtState);
  }
}

/*
** Check that local bus clock DCM has achieved lock.
*/
static bool_t
waitForLclkDcm(
  Globals* pGlob)
{
  uint32_t status;
  uint32_t val32;

  /*
  ** Delay for 500ms to allow DLLs/DCMs time to lock, and then verify
  ** that they are locked as expected.
  */
  sleepMs(500);
  status = pGlob->pFpgaRegs[STATUS_REG];
  _tprintf(_T("STATUS=0x%08lX\n"), (unsigned long)status);
  if (!(status & STATUS_REG_LCLKLOCKED)) {
    _tprintf(_T("*** LCLK DCM not locked\n"));
    return FALSE;
  }
  /* Clear "sticky" loss of lock bit */
  val32 = STATUS_REG_LCLKSTICKY;
  pGlob->pFpgaRegs[STATUS_REG] = val32;
  pGlob->pFpgaRegs[STATUS_REG];

  return TRUE;
}

/*
** Check that DCMs/PLLs in memory clock domain have achieved lock, and that
** any memory banks that are self-training have trained themselves without
** error.
*/
static bool_t
checkForMemoryLock(
  Globals* pGlob,
  unsigned int numLockFlags)
{
  uint32_t status, memstat, mask, val32;

  mask = (1 << numLockFlags) - 1;
  status = pGlob->pFpgaRegs[STATUS_REG];
  memstat = pGlob->pFpgaRegs[MEMSTAT_REG];
  _tprintf(_T("STATUS=0x%08lX MEMSTAT=0x%08lX\n"), (unsigned long)status, (unsigned long)memstat);
  val32 = status >> STATUS_REG_SHIFT_LOCKED;
  val32 &= mask;
  if (val32 != mask) {
    _tprintf(_T("*** Not all lock flags asserted: expected=0x%lX actual=0x%lX\n"),
      (unsigned long)mask, (unsigned long)val32);
    return FALSE;
  }
  val32 = memstat >> MEMSTAT_REG_SHIFT_TRAINED;
  val32 &= pGlob->bankTestMask;
  if (val32 != pGlob->bankTestMask) {
    _tprintf(_T("*** Not all memory banks trained: expected=0x%lX actual=0x%lX\n"),
      (unsigned long)pGlob->bankTestMask, (unsigned long)val32);
    return FALSE;
  }
  /* Clear sticky loss-of-lock / loss-of-trained bits */
  val32 = 0xffU << STATUS_REG_SHIFT_STICKY;
  pGlob->pFpgaRegs[STATUS_REG] = val32;
  pGlob->pFpgaRegs[STATUS_REG];

  return TRUE;
}

/*
** Check that DLLs/DCMs/PLLs in memory clock domain did not lose lock during
** the test, and that memory banks did not transition to untrained.
*/
static bool_t
checkStickyStatus(
  Globals* pGlob,
  unsigned int numLockFlags)
{
  uint32_t status, memstat, mask, val32;

  mask = (1 << numLockFlags) - 1;
  status = pGlob->pFpgaRegs[STATUS_REG];
  memstat = pGlob->pFpgaRegs[MEMSTAT_REG];
  _tprintf(_T("STATUS=0x%08lX MEMSTAT=0x%08lX\n"), (unsigned long)status, (unsigned long)memstat);
  val32 = status >> STATUS_REG_SHIFT_STICKY;
  val32 &= mask;
  if (val32) {
    _tprintf(_T("*** Loss of lock occurred: flags=0x%lX\n"), (unsigned long)val32);
    return FALSE;
  }

  return TRUE;
}

static uint64_t
verifyChunk(
  Globals* pGlob,
  uint8_t maskWord[MAX_BANK_WIDTH],   /* Mask word indicating which bits of a memory word should be checked */
  uint8_t* pBufExpected,              /* Buffer holding expected data */
  unsigned int wordWidth,             /* Width of memory word, in bytes */
  uint64_t location,                  /* Word-addressed memory location corresponding to start of pBufChk */
  size_t n,                           /* Size of expect / actual buffers, in bytes */
  uint64_t* pTotErr,					/* Total number of errors so far; updated on return */
  bool_t bUseDma,						/* TRUE => use DMA to read back data */
  bool_t bRetryReads)                   /* TRUE => retry reads */
{
  uint8_t* pBufChk = pGlob->pBufChk;
  uint64_t totErr;
  unsigned int maxRetry = 10;
  unsigned long pos;
  size_t chunk = n;
  unsigned int nRetry;
  uint64_t offset;
  uint8_t expectedWord[MAX_BANK_WIDTH];
  uint8_t actualWord[MAX_BANK_WIDTH];
  bool_t bOk = TRUE;

  totErr = *pTotErr;
  pos = 0;
  offset = location * wordWidth;
  readMemory(pGlob, pBufChk + pos, offset + pos, chunk, bUseDma);
  while (pos < chunk) {
    bOk = TRUE;
    for (; pos < chunk;) {
      maskExpectedWord(expectedWord, pBufExpected + pos, maskWord, wordWidth);
      if (verifyWord(pBufChk + pos, expectedWord, actualWord, maskWord, wordWidth)) {
        bOk = FALSE;
        totErr++;
        displayError(location, expectedWord, actualWord, wordWidth, totErr);
        if (bRetryReads) {
          break;
        }
      }
      location++;
      pos += wordWidth;
    }
    if (!bOk && bRetryReads) {
      nRetry = maxRetry;
      _tprintf(_T("[Retrying read at location 0x%llX]\n"), (unsigned long long)location);
      while (nRetry) {
        readMemory(pGlob, pBufChk + pos, offset + pos, wordWidth, bUseDma);
        if (verifyWord(pBufChk + pos, expectedWord, actualWord, maskWord, wordWidth)) {
          totErr++;
          displayError(location, expectedWord, actualWord, wordWidth, totErr);
        } else {
          break;
        }
        nRetry--;
      }
      if (0 == nRetry) {
        _tprintf(_T("[Giving up on location 0x%llX]\n"), (unsigned long long)location);
      }
      location++;
      pos += wordWidth;
    }
  }
  *pTotErr = totErr;

  return totErr;
}

static bool_t
getDimmMode(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
  unsigned int bank,
  uint32_t* pMode,
  bool_t* x4)
{
  uint8_t spd[128];
  bool_t bPresent;
  unsigned long size;
  unsigned int rows, cols, banks, pbanks;
  unsigned int devWidth;
  uint8_t clMask;
  uint8_t modAttr;
  bool_t bRegistered;
  uint32_t mode = 0;

  bPresent = readSpd(pGlob, bank, spd);
  if (bPresent) {
    rows = spd[3];      /* Number of bits used for row address on devices */
    cols = spd[4];      /* Number of bits used for column address on devices */
    pbanks = spd[5];    /* Number of physical banks on DIMM */
    devWidth = spd[13]; /* DQ width of devices */
    banks = spd[17];    /* Number of banks on devices */
    clMask = spd[18];   /* CAS latencies supported by devices */
    modAttr = spd[21];  /* DIMM module attributes */
    switch (devWidth) {
    case 4:
    case 8:
    case 16:
      break;

    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): unknown device data width (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)devWidth);
      return FALSE;
    }
    if (banks < 1 || banks > 8) {
      _tprintf(_T(" *** Bank %lu (DDR SDRAM): unsupported device banks value (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    if (!(clMask & 0x8U)) {
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): CAS latency 2.5 not supported by devices, not testing\n"),
        (unsigned long)bank);
      return FALSE;
    }
    switch (modAttr) {
    case 0x20: /* Unbuffered */
      bRegistered = FALSE;
      break;
    case 0x26: /* Registered */
      bRegistered = TRUE;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid DIMM module attributes value (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)modAttr);
      return FALSE;
    }
    switch (rows) {
    case 12:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_12;
      break;
    case 13:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
      break;
    case 14:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_14;
      break;
    case 15:
      mode |= MODEx_REG_DDRSDRAM_ROWBITS_15;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of row bits (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)rows);
      return FALSE;
    }
    switch (cols + 4 - rows) {
    case 0:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
      break;
    case 1:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
      break;
    case 3:
      mode |= MODEx_REG_DDRSDRAM_COLBITS_M1;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of column bits (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)cols);
      return FALSE;
    }
    switch (banks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_0;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_1;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of banks (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)banks);
      return FALSE;
    }
    switch (pbanks) {
    case 1:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
      break;
    case 2:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_2;
      break;
    case 4:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_4;
      break;
    case 8:
      mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_8;
      break;
    default:
      _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid number of physical banks (0x%02lX), not testing\n"),
        (unsigned long)bank, (unsigned long)pbanks);
      return FALSE;
    }
    if (bRegistered) {
      mode |= 0x1U;
    }
    if (devWidth == 4) {
      mode |= 0x2U;
    }
    pBankInfo[bank].Size = (1 << (rows + cols - 1)) * banks * pbanks;
    size = (pBankInfo[bank].Size >> 20) * (pBankInfo[bank].Width >> 3);
    _tprintf(_T(" Bank %lu (DDR SDRAM): x%lu %s rows=%lu cols=%lu banks=%lu physBanks=%lu %lu MiB\n"),
      (unsigned long)bank,
      (unsigned long)devWidth,
      bRegistered ? "registered" : "unbuffered",
      (unsigned long)rows, (unsigned long)cols,
      (unsigned long)banks, (unsigned long)pbanks, (unsigned long)size);
  } else {
    _tprintf(_T(" +++ Bank %lu (DDR SDRAM): SPD data not found, not testing\n"), (unsigned long)bank);
    return FALSE;
  }

  /* Bank SPD seems OK */
  *pMode = mode;
  *x4 = (devWidth == 4);
  return TRUE;
}

static void
setMemoryPortConfiguration(
  Globals* pGlob,
  uint32_t* modes)
{
  unsigned int i;

  _tprintf(_T("Configuring memory ports...\n"));
  for (i = 0; i < MAX_BANK; i++) {
    if (!(pGlob->bankTestMask & (0x1UL << i))) {
      continue;
    }
    pGlob->pFpgaRegs[MODEx_REG(i)] = modes[i];
  }
}

static void
scanMemoryPorts(
  ADMXRC2_CARD_INFO* pCardInfo,
  ADMXRC2_BANK_INFO* pBankInfo,
  Globals* pGlob,
  uint32_t* modes)
{
  unsigned int i;
  bool_t dimmX4[4];

  /* Initialize memory bank config registers where necessary */
  _tprintf(_T("Scanning memory ports...\n"));
  for (i = 0; i < pCardInfo->NumRAMBank; i++) {
    if (!(pGlob->bankTestMask & (0x1U << i))) {
      continue;
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SRAM_DDR2) {
      /* Select BL4 for DDR-II SSRAM and ensure DLL is enabled */
      modes[i] = MODEx_REG_DDR2SSRAM_BL4 | MODEx_REG_DDR2SSRAM_DLLON;
      pGlob->bBankDdr[i] = TRUE;
      _tprintf(_T(" Bank %lu (DDR-II SRAM): burst length 4, DLL enabled\n"), (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTP) {
      /* Select pipelined mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_PIPELINE;
      pGlob->bBankDdr[i] = FALSE;
      _tprintf(_T(" Bank %lu (ZBT SSRAM): pipelined mode\n"), (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_ZBTFT) {
      /* Select flowthrough mode for ZBT SSRAM */
      modes[i] = MODEx_REG_ZBTSSRAM_FLOWTHRU;
      pGlob->bBankDdr[i] = FALSE;
      _tprintf(_T(" Bank %lu (ZBT SSRAM): flowthrough mode\n"), (unsigned long)i);
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR) {
      uint32_t mode;

      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXPL:
      case ADMXRC2_BOARD_ADMXP:
        mode = 0;
        if (0x800000 == pBankInfo[i].Size) {
          /* unbuffered, row bits 13, col bits 9, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col 9, bank 2, phys banks 1\n"), (unsigned long)i);
        } else if (0x1000000 == pBankInfo[i].Size) {
          /* unregistered, row bits 13, col bits 10, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col 10, bank 2, phys banks 1\n"), (unsigned long)i);
        } else if (0x2000000 == pBankInfo[i].Size) {
          /* unregistered, row bits 13, col bits 11, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M2;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col  11, bank 2, phys banks 1\n"), (unsigned long)i);
        } else {
          pGlob->bankTestMask &= ~(0x1U << i);
          _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid size (0x%lX words), not testing\n"),
            (unsigned long)i, (unsigned long)pBankInfo[i].Size);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;

      case ADMXRC2_BOARD_ADPXPI:
        if (!getDimmMode(pCardInfo, pBankInfo, pGlob, i, &mode, &dimmX4[i])) {
          pGlob->bankTestMask &= ~(0x1U << i);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;
#if ZAP
      case ADMXRC2_BOARD_ADMXRC4SX:
      case ADMXRC2_BOARD_ADMXRC4LX:
        mode = 0;
        if (0x1000000 == pBankInfo[i].Size) {
          /* unregistered, row bits 13, col bits 9, bank bits 2, phys banks 1 */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_2;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          _tprintf(_T(" Bank %lu (DDR SDRAM): unbuffered, row 13, col 10, bank 2, phys banks 1\n"), (unsigned long)i);
        } else {
          pGlob->bankTestMask &= ~(0x1U << i);
          _tprintf(_T(" +++ Bank %lu (DDR SDRAM): invalid size (0x%lX words), not testing\n"),
            (unsigned long)i, (unsigned long)pBankInfo[i].Size);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;
#endif /* ZAP */

      default:
        assert(FALSE); /* Shouldn't get here */
        break;
      }
    }
    if (pBankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR2) {
      uint32_t mode;

      /* Select registered, row bits, col bits, bank bits and phys bank bits for DDR SDRAM */
      switch (pCardInfo->BoardType) {
      case ADMXRC2_BOARD_ADMXRC4FX:
      case ADMXRC2_BOARD_ADPEXRC4FX:
      case ADMXRC2_BOARD_ADMXRC5LX:
      case ADMXRC2_BOARD_ADMXRC5T1:
      case ADMXRC2_BOARD_ADMXRC5T2:
      case ADMXRC2_BOARD_ADMXRC5T2ADV:
      case ADMXRC2_BOARD_ADMXRC5T2ADV6:
      case ADMXRC2_BOARD_ADMXRC5TDA1:
      case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
        mode = 0;
        if (0x1000000 == pBankInfo[i].Size) {
          /* unbuffered, row bits 13, col bits 10, bank bits 3, phys banks 1, Trfc for 1Gb part */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_13;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M3;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          mode |= MODEx_REG_DDR2SDRAM_TRFC_1Gb;
          _tprintf(_T(" Bank %lu (DDR-II SDRAM): unbuffered, row 13, col 10, bank 3, phys banks 1\n"),
            (unsigned long)i);
        } else if (0x2000000 == pBankInfo[i].Size) {
          /* unbuffered, row bits 14, col bits 10, bank bits 3, phys banks 1, Trfc for 2Gb part */
          mode |= MODEx_REG_DDRSDRAM_UNBUFFERED;
          mode |= MODEx_REG_DDRSDRAM_ROWBITS_14;
          mode |= MODEx_REG_DDRSDRAM_COLBITS_M4;
          mode |= MODEx_REG_DDRSDRAM_BANKBITS_3;
          mode |= MODEx_REG_DDRSDRAM_PHYSBANKS_1;
          mode |= MODEx_REG_DDR2SDRAM_TRFC_2Gb;
          _tprintf(_T(" Bank %lu (DDR-II SDRAM): unbuffered, row 14, col 10, bank 3, phys banks 1\n"),
            (unsigned long)i);
        } else {
          pGlob->bankTestMask &= ~(0x1U << i);
          _tprintf(_T(" +++ Bank %lu (DDR-II SDRAM): invalid size (0x%lX words), not testing\n"),
            (unsigned long)i, (unsigned long)pBankInfo[i].Size);
        }
        modes[i] = mode;
        pGlob->bBankDdr[i] = TRUE;
        break;

      default:
        assert(FALSE); /* Shouldn't get here */
        break;
      }
    }
  }
}

int
_tmain(
  int argc,
  TCHAR *argv[])
{
  Globals glob;
  int ret = 0;
  ADMXRC2_STATUS status;
  ADMXRC2_CARD_INFO cardInfo;
  ADMXRC2_SPACE_INFO spInfo;
  ADMXRC2_BANK_INFO bankInfo[MAX_BANK];
  const TCHAR* pFilename;
  TCHAR* pDesignName;
  Option options[12];
  Arguments arguments;
  unsigned int i, rep;
  unsigned int numLockFlags;
  int dma;
  uint64_t memorySize;
  bool_t bUseDma;
  bool_t bUsePio;
  bool_t bMeasurePerf;
  bool_t bSetMclk, bAllow64Bit;
  uint32_t modes[MAX_BANK];
  double lclkFreq, actualLclkFreq;
  double mclkFreq, actualMclkFreq;
  unsigned long nrep;
  bool_t bFailed;
  uint8_t mask[MAX_BANK_WIDTH];
  bool_t bUseMask;
  bool_t bRetryReads;

  /***********************************************************************
  ** 1. Initialize stuff so that we know where we are if we have to     **
  **    clean up if some later step fails.                              **
  ***********************************************************************/

  for (i = 0; i < 0x100; i++) {
    unsigned int j;
    uint8_t val;

    val = 0;
    for (j = 0; j < 8; j++) {
      val |= ((i >> j) & 0x1U) << (7 - j);
    }
    glob.revLookup[i] = val;
  }
  glob.lastSpinTick = timeGet();
  glob.spIndex = 0;
  glob.pBuf5a = NULL;
  glob.pBufOwnAddr = NULL;
  glob.pBufRevAddr = NULL;
  glob.pBufRandom = NULL;
  glob.pBufChk = NULL;
  glob.hCard = ADMXRC2_HANDLE_INVALID_VALUE;
  glob.hDmaEvent = NULL;

  /***********************************************************************
  ** 2. Set up the data structure that defines command-line syntax.     **
  ***********************************************************************/

  options[0].key = _T("banks");
  options[0].type = Option_hex;
  options[0].def.hexVal = 0xffffU;
  options[0].meaning = _T("bitmask of banks to test");
  options[1].key = _T("card");
  options[1].type = Option_uint;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("ID of card to use");
  options[2].key = _T("index");
  options[2].type = Option_uint;
  options[2].def.uintVal = 0;
  options[2].meaning = _T("index of card to use");
  options[3].key = _T("repeat");
  options[3].type = Option_uint;
  options[3].def.uintVal = 1;
  options[3].meaning = _T("number of repetitions");
  options[4].key = _T("perf");
  options[4].type = Option_boolean;
  options[4].def.booleanVal = TRUE;
  options[4].meaning = _T("TRUE => measure DMA-to-bank performance");
  options[5].key = _T("usedma");
  options[5].type = Option_boolean;
  options[5].def.booleanVal = TRUE;
  options[5].meaning = _T("TRUE => test using DMA");
  options[6].key = _T("usepio");
  options[6].type = Option_boolean;
  options[6].def.booleanVal = FALSE;
  options[6].meaning = _T("TRUE => test using programmed I/O");
  options[7].key = _T("64");
  options[7].type = Option_boolean;
  options[7].def.booleanVal = FALSE;
  options[7].meaning = _T("TRUE => use 64 bit local bus");
  options[8].key = _T("lclk");
  options[8].type = Option_float;
  options[8].def.floatVal = 33.0;
  options[8].meaning = _T("local bus clock frequency, in MHz");
  options[9].key = _T("mclk");
  options[9].type = Option_float;
  options[9].def.floatVal = 33.0;
  options[9].meaning = _T("MCLK frequency, in MHz");
  options[10].key = _T("mask");
  options[10].type = Option_string;
  options[10].def.stringVal = _T("<none>");
  options[10].meaning = _T("physical bit lanes to test");
  options[11].key = _T("retry");
  options[11].type = Option_boolean;
  options[11].def.booleanVal = FALSE;
  options[11].meaning = _T("TRUE => retry reads on verification error");
  arguments.nOption = 12;
  arguments.option = options;
  arguments.nParamType = 0;
  arguments.minNParam = 0;

  /***********************************************************************
  ** 3. Parse the command line                                          **
  ***********************************************************************/

  parse_command_line(_T("MEMORY example application"), _T("memory"), argc, argv, &arguments);

  if (options[1].specified && options[2].specified) {
    _tprintf(_T("*** Do not specify both -card and -index\n"));
    ret = -1;
    goto done;
  }

  nrep = options[3].value.uintVal;
  bMeasurePerf = options[4].value.booleanVal;
  bUseDma = options[5].value.booleanVal;
  bUsePio = options[6].value.booleanVal;
  glob.bUse64Bit = options[7].value.booleanVal;
  bRetryReads = options[11].value.booleanVal;

  /***********************************************************************
  ** 4. Open the selected card                                          **
  ***********************************************************************/

  if (!options[2].specified) {
    ADMXRC2_CARDID cardID = options[1].value.uintVal;

    status = ADMXRC2_OpenCard(cardID, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to open card with ID %lu: %s\n"),
        (unsigned long)cardID, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  } else {
    unsigned int index = options[2].value.uintVal;

    status = ADMXRC2_OpenCardByIndex(index, &glob.hCard);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
        (unsigned long)index, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  /***********************************************************************
  ** 5. Get card information for the selected card. This will be used   **
  **    later to determine default clock frequencies and other model-   **
  **    specific settings.                                              **
  ***********************************************************************/

  status = ADMXRC2_GetCardInfo(glob.hCard, &cardInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to get card info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 6. Figure out model-specific settings from card type.              **
  ***********************************************************************/

  switch (cardInfo.BoardType) {
    case ADMXRC2_BOARD_ADMXRC:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 33.33;
      mclkFreq = 40.0;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC_P:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 33.33;
      mclkFreq = 40.0;
      numLockFlags = 1;
      break;

    case ADMXRC2_BOARD_ADMXRC2_LITE:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 40.0;
      mclkFreq = 66.67;
      numLockFlags = 3;
      break;

    case ADMXRC2_BOARD_ADMXRC2:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 50.0;
      mclkFreq = 66.67;
      numLockFlags = 3;
      break;

#if ZAP
    case ADMXRC2_BOARD_ADPDRC2:
      {
        uint32_t ctlRegsLocalAddr;
        uint32_t pconAddr;
        uint8_t pcon;

        bAllow64Bit = FALSE;
        bSetMclk = TRUE;
        lclkFreq = 66.67;
        mclkFreq = 66.67;
        numLockFlags = 2;
        //            design = glob.bUse64Bit ? "memory64_x8" : "memory_x8");

        /* Get the local bus address of CPLD registers */
        status = ADMXRC2_GetSpaceInfo(glob.hCard, 1, &spInfo);
        if (status != ADMXRC2_SUCCESS) {
          _tprintf(_T("*** Failed to get space 1 info: %s\n",
            ADMXRC2_GetStatusString(status));
          ret = -1;
          goto done;
        }
        ctlRegsLocalAddr = spInfo.LocalBase;

        /* Give FPGA access to I2C bus */
        pconAddr = ctlRegsLocalAddr + 0x5U;
        status = ADMXRC2_Read(glob.hCard, ADMXRC2_IOWIDTH_8, 0, pconAddr, &pcon, 1);
        if (status != ADMXRC2_SUCCESS) {
          _tprintf(_T("Failed to read PCON register: %s\n",
            ADMXRC2_GetStatusString(status));
          ret = -1;
          goto done;
        }
        pcon |= 0x80;
        status = ADMXRC2_Write(glob.hCard, ADMXRC2_IOWIDTH_8, 0, pconAddr, &pcon, 1);
        if (status != ADMXRC2_SUCCESS) {
          _tprintf(_T("Failed to write PCON register: %s\n",
            ADMXRC2_GetStatusString(status));
          ret = -1;
          goto done;
        }
        status = ADMXRC2_Read(glob.hCard, ADMXRC2_IOWIDTH_8, 0, pconAddr, &pcon, 1);
        if (status != ADMXRC2_SUCCESS) {
          _tprintf(_T("Failed to read PCON register: %s\n",
            ADMXRC2_GetStatusString(status));
          ret = -1;
          goto done;
        }
      }
      break;

#endif /* ZAP */
#if ZAP
    case ADMXRC2_BOARD_ADPWRC2:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      mclkFreq = 58.333;
      numLockFlags = 2;
      break;

#endif /* ZAP */
    case ADMXRC2_BOARD_ADMXPL:
      {
        bAllow64Bit = TRUE;
        bSetMclk = FALSE;
        lclkFreq = 66.67;
        mclkFreq = 2.0 * lclkFreq;
        numLockFlags = 2;
      }
      break;

    case ADMXRC2_BOARD_ADMXP:
      {
        bAllow64Bit = TRUE;
        bSetMclk = FALSE;
        lclkFreq = 66.67;
        mclkFreq = 2.0 * lclkFreq;
        numLockFlags = 1;
      }
      break;

    case ADMXRC2_BOARD_ADPXPI:
      {
        bAllow64Bit = TRUE;
        bSetMclk = FALSE;
        lclkFreq = 66.67;
        mclkFreq = 2.0 * lclkFreq;
        numLockFlags = 1;
      }
      break;

    case ADMXRC2_BOARD_ADMXRC4LS:
    case ADMXRC2_BOARD_ADMXRC4LX:
    case ADMXRC2_BOARD_ADMXRC4SX:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 66.0;
      mclkFreq = 150.0; /* 150MHz is max 1X output frequency for DCM in low frequency mode */
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC4FX:
      {
        bAllow64Bit = TRUE;
        bSetMclk = TRUE;
        lclkFreq = 66.67;
        /* 200.0MHz MCLK => 100.0MHz in memory clock domain, 200.0MHz DDR-II SDRAM clock */
        mclkFreq = 200.0;
        numLockFlags = 2;
      }
      break;

    case ADMXRC2_BOARD_ADPEXRC4FX:
      {
        bAllow64Bit = TRUE;
        bSetMclk = TRUE;
        lclkFreq = 66.67;
        /* 200.0MHz MCLK => 100.0MHz in memory clock domain, 200.0MHz DDR-II SDRAM clock */
        mclkFreq = 200.0;
        numLockFlags = 4;
      }
      break;

    case ADMXRC2_BOARD_ADMXRC5LX:
      {
        bAllow64Bit = TRUE;
        bSetMclk = TRUE;
        lclkFreq = 66.67;
        /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
        mclkFreq = 266.67;
        numLockFlags = 2;
      }
      break;

    case ADMXRC2_BOARD_ADMXRC5T1:
      {
        bAllow64Bit = TRUE;
        bSetMclk = TRUE;
        lclkFreq = 66.67;
        /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
        mclkFreq = 266.67;
        numLockFlags = 2;
      }
      break;

    case ADMXRC2_BOARD_ADMXRC5T2:
    case ADMXRC2_BOARD_ADMXRC5T2ADV:
    case ADMXRC2_BOARD_ADMXRC5T2ADV6:
    case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
      {
        bAllow64Bit = TRUE;
        bSetMclk = TRUE;
        lclkFreq = 66.67;
        /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
        mclkFreq = 266.67;
        numLockFlags = 2;
      }
      break;

    case ADMXRC2_BOARD_ADMXRC5TZ:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 166.67MHz MCLK => 166.67MHz in memory clock domain, 166.67MHz ZBT SSRAM clock */
      mclkFreq = 166.67;
      numLockFlags = 2;
      break;

    case ADMXRC2_BOARD_ADMXRC5TDA1:
      bAllow64Bit = FALSE;
      bSetMclk = TRUE;
      lclkFreq = 66.67;
      /* 266.67MHz MCLK => 133.33MHz in memory clock domain, 266.67MHz DDR-II SDRAM clock */
      mclkFreq = 266.67;
      numLockFlags = 2;
      break;

    default:
      _tprintf(_T("*** This example must be run on one of the following cards:\n\n"));
      _tprintf(_T("\tADM-XRC\n"));
      _tprintf(_T("\tADM-XRC-P\n"));
      _tprintf(_T("\tADM-XRC-II-Lite\n"));
      _tprintf(_T("\tADM-XRC-II\n"));
      _tprintf(_T("\tADM-XPL\n"));
      _tprintf(_T("\tADM-XP\n"));
      _tprintf(_T("\tADP-WRC-II\n"));
      _tprintf(_T("\tADP-DRC-II\n"));
      _tprintf(_T("\tADP-XPI\n"));
      _tprintf(_T("\tADM-XRC-4LX\n"));
      _tprintf(_T("\tADM-XRC-4SX\n"));
      _tprintf(_T("\tADM-XRC-4FX / ADM-XMC-4FX\n"));
      _tprintf(_T("\tADPE-XRC-4FX\n"));
      _tprintf(_T("\tADM-XRC-5LX\n"));
      _tprintf(_T("\tADM-XRC-5T1\n"));
      _tprintf(_T("\tADM-XRC-5T2 / ADM-XRC-5T2-ADV / ADM-XRC-5T2-ADV6\n"));
      _tprintf(_T("\tADM-XRC-5TZ\n"));
      _tprintf(_T("\tADM-XRC-5T-DA1\n"));
      _tprintf(_T("\tADM-XRC-5T2-ADV-CC1\n"));
      ret = -1;
      goto done;
  }

  if (glob.bUse64Bit && !bAllow64Bit) {
    _tprintf(_T("*** 64-bit local bus is not supported on the specified card.\n"));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 7. Get the addresses for the FPGA registers and the memory access  **
  **    window, in our virtual address space.                           **
  ***********************************************************************/

  status = ADMXRC2_GetSpaceInfo(glob.hCard, 0, &spInfo);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to get space 0 info: %s\n"), ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }
  glob.ppFpgaSpace = spInfo.VirtualBase;
  if (glob.ppFpgaSpace == NULL) {
    _tprintf(_T("*** Failed to retrieve pointer to FPGA space\n"));
    ret = -1;
    goto done;
  }
  glob.pFpgaRegs = (volatile uint32_t*)glob.ppFpgaSpace;
  glob.pMemoryWindow = (uint8_t*)glob.ppFpgaSpace;
  glob.pMemoryWindow += memWindowAddr;


  /***********************************************************************
  ** 8. Get memory bank information.                                    **
  **                                                                    **
  **    To determine the byte width of a memory bank, we first divide   **
  **    by 9. If there is no remainder, then that bank has parity bits, **
  **    and we have calculated the byte width. If there is a remainder, **
  **    then we assume the bank has no parity bits, in which case we    **
  **    should be able to calculate the byte width by dividing by 8.    **
  **                                                                    **
  **    For the ADM-XRC-II-Lite, we fudge the data a little. Each pair  **
  **    of 18-bit wide banks is combined to get a logical 32-bit wide   **
  **    bank.                                                           **
  ***********************************************************************/

  /* Get bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    if (cardInfo.RAMBanksFitted & (0x1UL << i)) {
      status = ADMXRC2_GetBankInfo(glob.hCard, i, &bankInfo[i]);
      if (status != ADMXRC2_SUCCESS) {
        _tprintf(_T("*** Failed to get bank %lu info: %s\n"), (unsigned long)i,
          ADMXRC2_GetStatusString(status));
        ret = -1;
        goto done;
      }

      if (bankInfo[i].Type & ADMXRC2_RAM_SDRAM_DDR2) {
        /*
        ** The DDR-II SSRAM memory interface for Virtex 4/5 has a user-
        ** interface that is 4 times as wide as the physical interface
        ** to the memory device.
        */
        bankInfo[i].Size >>= 1;
        bankInfo[i].Width <<= 1;
      }

      if (bankInfo[i].Type & ADMXRC2_RAM_SRAM_DDR2) {
        switch (cardInfo.BoardType) {
        case ADMXRC2_BOARD_ADPEXRC4FX:
        case ADMXRC2_BOARD_ADMXRC5T1:
        case ADMXRC2_BOARD_ADMXRC5T2:
        case ADMXRC2_BOARD_ADMXRC5T2ADV:
        case ADMXRC2_BOARD_ADMXRC5T2ADV6:
        case ADMXRC2_BOARD_ADMXRC5TDA1:
        case ADMXRC2_BOARD_ADMXRC5T2ADVCC1:
          /*
          ** The high speed DDR-II SSRAM memory interface for Virtex 4/5
          ** has a user-interface that is 4 times as wide as the
          ** physical interface to the memory device.
          */
          bankInfo[i].Size >>= 1;
          bankInfo[i].Width <<= 1;
          break;

        default:
          assert(FALSE); /* Shouldn't get here */
          break;
        }
      }
    }
  }

  /*
  ** For the ADM-XRC-II-Lite, fix up the memory bank information to make
  ** it look like we have 32-bit wide banks (we combine two 18-bit physical
  ** banks into one logical bank).
  */
  if (cardInfo.BoardType == ADMXRC2_BOARD_ADMXRC2_LITE) {
    uint32_t banksFitted = 0x0U;

    cardInfo.NumRAMBank = 2;
    for (i = 0; i < cardInfo.NumRAMBank; i++) {
      unsigned int j = i * 2;
      bool_t bFittedA, bFittedB;
      bool_t bSame;

      bFittedA = (cardInfo.RAMBanksFitted >> j) != 0;
      bFittedB = (cardInfo.RAMBanksFitted >> (j + 1)) != 0;
      if (bFittedA && bFittedB) {
        bSame = (bankInfo[j].Width == bankInfo[j + 1].Width) &&
          (bankInfo[j].Type == bankInfo[j + 1].Type);
        if (bSame) {
          bankInfo[i].Size = minimumOf2Macro(bankInfo[j].Size, bankInfo[j + 1].Size);
          bankInfo[i].Type = bankInfo[j].Type & bankInfo[j + 1].Type;
          bankInfo[i].Fitted = bankInfo[j].Fitted & bankInfo[j + 1].Fitted;
          banksFitted |= (0x1U << i);
        } else {
          /* This logical bank has been deemed invalid */
          _tprintf(_T("+++ Can't pair banks %lu and %lu due to different configuration\n"),
            (unsigned long)j, (unsigned long)(j + 1));
        }
      }
    }
    cardInfo.RAMBanksFitted = banksFitted;
  }

  /*
  ** Establish the bank test mask. This is the set of banks that are we
  ** have been instructed to test AND are fitted. This value may be
  ** further modified during the further stages of this application in
  ** order to remove banks that are deemed invalid.
  */
  glob.bankTestMask = options[0].value.hexVal & cardInfo.RAMBanksFitted;

  /* Process bank information */
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;

    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;

    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }

    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }

    if ((bankInfo[i].Width % 9) == 0) {
      /* Bank has parity bits - don't use to calculate bank size */
      glob.bankWidth[i] = bankInfo[i].Width / 9;
    } else {
      /* Bank doesn't have parity bits */
      glob.bankWidth[i] = bankInfo[i].Width / 8;
    }
  }

  /***********************************************************************
  ** 9. Parse any mask that might have been specified on the command-   **
  **    line. This may be useful for diagnosing certain types of memory **
  **    fault like a stuck data bit or a short to another data bit. The **
  **    mask specifies which bits of a memory word are actually         **
  **    verified during the memory tests.                               **
  ***********************************************************************/

  memset(mask, 0xffU, MAX_BANK_WIDTH);
  if (options[10].specified) {
    size_t i, n;
    uint8_t digit;
    unsigned int shift;
    TCHAR c;

    bUseMask = TRUE;
    n = _tcslen(options[10].value.stringVal);
    if (n < 1) {
      _tprintf(_T("*** Invalid value for -mask option: %s\n"), options[10].value.stringVal);
      ret = -1;
      goto done;
    }
    memset(mask, 0x0U, MAX_BANK_WIDTH);
    for (i = 0; i < n; i++) {
      c = _totlower(options[10].value.stringVal[i]);
      if (c >= _T('a') && c <= _T('f')) {
        digit = c - 'a' + 10;
      } else if (c >= _T('0') && c <= _T('9')) {
        digit = c - _T('0');
      } else {
        _tprintf(_T("*** Invalid digit %lu in value for -mask option: %s\n"),
          (unsigned long)(i + 1), options[10].value.stringVal);
        ret = -1;
        goto done;
      }
      shift = ((n - i - 1) & 0x1U) << 2;
      mask[(n - i - 1) >> 1] |= digit << shift;
    }
  } else {
    bUseMask = FALSE;
  }

  /***********************************************************************
  ** 10. Allocate the buffers that are used to hold test data patterns  **
  **     and the data read back during the memory tests.                **
  ***********************************************************************/

  glob.pBuf5a = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBuf5a == NULL) {
    _tprintf(_T("*** Failed to allocate 0x5A buffer - aborting\n"));
    ret = -1;
    goto done;
  }

  glob.pBufOwnAddr = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufOwnAddr == NULL) {
    _tprintf(_T("*** Failed to allocate own address buffer - aborting\n"));
    ret = -1;
    goto done;
  }

  glob.pBufRevAddr = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufRevAddr == NULL) {
    _tprintf(_T("*** Failed to allocate own address buffer - aborting\n"));
    ret = -1;
    goto done;
  }

  glob.pBufRandom = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufRandom == NULL) {
    _tprintf(_T("*** Failed to allocate random data buffer - aborting\n"));
    ret = -1;
    goto done;
  }

  glob.pBufChk = (void*)malloc(GENERATE_CHUNK);
  if (glob.pBufChk == NULL) {
    _tprintf(_T("*** Failed to allocate RAM readback buffer - aborting\n"));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 11. Program the frequencies for LCLK and MCLK. The defaults are    **
  **     obtained from the board type, but may be overridden on the     **
  **     command-line.                                                  **
  ***********************************************************************/

  if (options[8].specified) {
    lclkFreq = options[8].value.floatVal;
  }

  if (options[9].specified) {
    mclkFreq = options[9].value.floatVal;
  }

  _tprintf(_T("Setting LCLK frequency to %.1fMHz\n"), lclkFreq);
  status = ADMXRC2_SetClockRate(glob.hCard, ADMXRC2_CLOCK_LCLK, lclkFreq * 1.0e6, &actualLclkFreq);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to set LCLK to %.1fMHz: %s\n"),
      lclkFreq, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  if (bSetMclk) {
    _tprintf(_T("Setting MCLK frequency to %.1fMHz\n"), mclkFreq);
    status = ADMXRC2_SetClockRate(glob.hCard, 1, mclkFreq * 1.0e6, &actualMclkFreq);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to set clock 1 to %.1fMHz: %s\n"),
        mclkFreq, ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
    _tprintf(_T("Actual MCLK frequency %.1fMHz\n"), actualMclkFreq / 1.0e6);
  }

  /***********************************************************************
  ** 12. Configure the FPGA with the "memory" bitstream.                **
  ***********************************************************************/

  pDesignName = glob.bUse64Bit ? _T("memory64") : _T("memory");
  pFilename = get_fpga_filename_dir(&cardInfo, pDesignName, pDesignName, &arguments);
  _tprintf(_T("Configuring FPGA with '%s'...\n"), pFilename);
  status = ADMXRC2_ConfigureFromFile(glob.hCard, pFilename);
  if (status != ADMXRC2_SUCCESS) {
    _tprintf(_T("*** Failed to load the bitstream '%s': %s\n"),
      pFilename, ADMXRC2_GetStatusString(status));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 13. After configuring the FPGA, it's necessary to wait a while     **
  **     to enable any DLLs, DCMs and PLLs that depend upon LCLK to     **
  **     achieve lock. The delay is a conservative 500ms, because we    **
  **     make no assumptions about whether or not an ES (engineering    **
  **     silicon) FPGA is being used.                                   **
  ***********************************************************************/

  if (!waitForLclkDcm(&glob)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 14. If running in 64-bit mode was requested on the command-line,   **
  **     configure Direct Slave space 0 for a 64-bit local bus. Also    **
  **     generate the DMA mode word for DMA using a 64-bit local bus.   **
  ***********************************************************************/

  if (glob.bUse64Bit) {
    uint32_t config = 0;

    config |= ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_WIDTH_64;
    status = ADMXRC2_SetSpaceConfig(glob.hCard, 0, config);
    if (status != ADMXRC2_SUCCESS) {
      _tprintf(_T("*** Failed to set space configuration: %s\n"), ADMXRC2_GetStatusString(status));
      ret = -1;
      goto done;
    }
  }

  glob.dmaMode = ADMXRC2_BuildDMAModeWord(
    cardInfo.BoardType,
    glob.bUse64Bit ? ADMXRC2_IOWIDTH_64 : ADMXRC2_IOWIDTH_32,
    0,
    ADMXRC2_DMAMODE_USEREADY | ADMXRC2_DMAMODE_USEBTERM | ADMXRC2_DMAMODE_BURSTENABLE);
#if defined(_WIN32)
  glob.hDmaEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
#endif

  /***********************************************************************
  ** 15. Here we use the memory bank information read in step 8 to      **
  **     determine the proper values for the MODE registers.            **
  ***********************************************************************/

  scanMemoryPorts(
    &cardInfo,
    bankInfo,
    &glob,
    modes);

  /***********************************************************************
  ** 16. Establish the MODE registers, based on what was determined     **
  **     in step 15.                                                    **
  ***********************************************************************/

  setMemoryPortConfiguration(&glob, modes);

  /***********************************************************************
  ** 17. Reset the memory subsystem, and wait for DLLs, DCMs, PLL and   **
  **     IDELAYCTRLs to achieve lock, as well as allowing any self-     **
  **     training memory interfaces sufficient time to train            **
  **     themselves.                                                    **
  ***********************************************************************/

  /*
  ** Reset the memory subsystem to ensure that memory banks can
  ** train themselves etc.
  */
  _tprintf(_T("Resetting memory subsystem...\n"));
  glob.pFpgaRegs[MEMCTL_REG] = 0x1U;
  glob.pFpgaRegs[MEMCTL_REG];
  sleepMs(200);
  glob.pFpgaRegs[MEMCTL_REG] = 0x0U;
  glob.pFpgaRegs[MEMCTL_REG];
  sleepMs(300);

  /* Check for memory DLL/DCM/PLL locking and training */
  if (!checkForMemoryLock(&glob, numLockFlags)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 18. Figure out total size (in bytes) of memory banks that are to   **
  **     be tested.                                                     **
  ***********************************************************************/

  memorySize = 0;
  for (i = 0; i < cardInfo.NumRAMBank && i < MAX_BANK; i++) {
    uint32_t typeMask = 0;

    typeMask |= ADMXRC2_RAM_ZBTFT;
    typeMask |= ADMXRC2_RAM_ZBTP;
    typeMask |= ADMXRC2_RAM_SRAM_DDR2;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR;
    typeMask |= ADMXRC2_RAM_SDRAM_DDR2;

    if (!(glob.bankTestMask & (0x1UL << i))) {
      /* Ignore banks that we're instructed not to test */
      continue;
    }

    if (!bankInfo[i].Fitted) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }

    if (!(bankInfo[i].Type & typeMask)) {
      /* Banks that aren't fitted are invalid */
      glob.bankTestMask &= ~(0x1U << i);
      continue;
    }

    glob.bankSize[i] = glob.bankWidth[i] * bankInfo[i].Size;
    memorySize += (glob.bankSize[i] >> 10);
  }

  /* If there's nothing to do, abort! */
  if (!glob.bankTestMask) {
    _tprintf(_T("*** Nothing to test - aborting.\n"));
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 19. This is where the memory tests begin. There are four different **
  **     tests:                                                         **
  **                                                                    **
  **      a) 55 and AA pattern, for detecting stuck / shorted data bits **
  **      b) Own address, for detecing faulty address lines             **
  **      c) Own address, bit reversed, for detecting faulty address    **
  **         lines                                                      **
  **      d) Randomly generated data, for detecting pattern-sensitive   **
  **         faults.                                                    **
  ***********************************************************************/

  _tprintf(_T("Total memory size to test : %lu kiB (0x%lX kiB)\n"),
    (unsigned long)memorySize, (unsigned long)memorySize);

  bFailed = FALSE;
  for (rep = 0; rep < nrep; rep++) {
    uint64_t nErr;
    unsigned long chunk;

    for (dma = 0; dma < 2; dma++) {
      uint64_t remaining, offset;
      uint64_t location;
      unsigned int width;
      TCHAR maskStr[1024];

      if (!dma && !bUsePio) {
        continue;
      }
      if (dma && !bUseDma) {
        continue;
      }

      _tprintf(_T("Performing memory tests using %s\n"), dma ? _T("DMA") : _T("programmed I/O"));

      _tcscpy_s(maskStr, ARRAY_LENGTH(maskStr), _T("         mask="));

      _tprintf(_T(" 5/A pattern...\n"));
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
        if (!(glob.bankTestMask & (0x1U << i))) {
          continue;
        }

        if (bUseMask) {
          displayMask(glob.bankWidth[i], mask, maskStr + 14, ARRAY_LENGTH(maskStr - 14));
        } else {
          maskStr[0] = _T('\0');
        }

        _tprintf(_T("   Bank %lu%s\r"), (unsigned long)i, maskStr);
        selectBank(&glob, i);
        width = glob.bankWidth[i];
        generate5A((uint8_t*)glob.pBuf5a, GENERATE_CHUNK, width, glob.bBankDdr[i]);

        /* Write data to memory bank */
        remaining = glob.bankSize[i];
        offset = 0;
        while (remaining) {
          chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
          writeMemory(&glob, glob.pBuf5a, offset, chunk, dma != 0);
          remaining -= chunk;
          offset += chunk;
          spin(&glob, TRUE);
        }

        /* Read back data from memory bank and check */
        remaining = glob.bankSize[i];
        offset = 0;
        location = 0;
        nErr = 0;
        while (remaining) {
          chunk = (GENERATE_CHUNK_RD > remaining) ? (unsigned long)remaining : GENERATE_CHUNK_RD;
          verifyChunk(
            &glob,
            mask,
            glob.pBuf5a,
            width,
            location,
            chunk,
            &nErr,
            dma ? TRUE : FALSE,
            bRetryReads);
          location += (chunk / width);
          offset += chunk;
          remaining -= chunk;
          spin(&glob, FALSE);
        }
        if (nErr) {
          _tprintf(_T("*** %lld(0x%llX) error(s) encountered while verifying data\n"),
            (unsigned long long)nErr, (unsigned long long)nErr);
          bFailed = TRUE;
          break;
        }
      }

      _tprintf(_T(" Own address...\n"));
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
        if (!(glob.bankTestMask & (0x1U << i))) {
          continue;
        }

        if (bUseMask) {
          displayMask(glob.bankWidth[i], mask, maskStr + 14, ARRAY_LENGTH(maskStr) - 14);
        } else {
          maskStr[0] = _T('\0');
        }

        _tprintf(_T("   Bank %lu%s\r"), (unsigned long)i, maskStr);
        selectBank(&glob, i);
        width = glob.bankWidth[i];

        /* Write data to memory bank */
        remaining = glob.bankSize[i];
        offset = 0;
        while (remaining) {
          chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
          generateOwnAddr((uint32_t*)glob.pBufOwnAddr, (uint32_t)(offset >> 2), chunk);
          writeMemory(&glob, glob.pBufOwnAddr, offset, chunk, dma);
          remaining -= chunk;
          offset += chunk;
          spin(&glob, TRUE);
        }

        /* Read back data from memory bank and check */
        remaining = glob.bankSize[i];
        offset = 0;
        location = 0;
        nErr = 0;
        while (remaining) {
          chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
          generateOwnAddr((uint32_t*)glob.pBufOwnAddr, (uint32_t)(offset >> 2), chunk);
          verifyChunk(
            &glob,
            mask,
            glob.pBufOwnAddr,
            width,
            location,
            chunk,
            &nErr,
            dma ? TRUE : FALSE,
            bRetryReads);
          location += (chunk / width);
          offset += chunk;
          remaining -= chunk;
          spin(&glob, FALSE);
        }
        if (nErr) {
          _tprintf(_T("*** %lld(0x%llX) error(s) encountered while verifying data\n"),
            (unsigned long long)nErr, (unsigned long long)nErr);
          bFailed = TRUE;
          break;
        }
      }

      _tprintf(_T(" Reversed own address...\n"));
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
        if (!(glob.bankTestMask & (0x1U << i))) {
          continue;
        }

        if (bUseMask) {
          displayMask(glob.bankWidth[i], mask, maskStr + 14, ARRAY_LENGTH(maskStr) - 14);
        } else {
          maskStr[0] = _T('\0');
        }

        _tprintf(_T("   Bank %lu%s\r"), (unsigned long)i, maskStr);
        selectBank(&glob, i);
        width = glob.bankWidth[i];

        /* Write data to memory bank */
        remaining = glob.bankSize[i];
        offset = 0;
        while (remaining) {
          chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
          generateRevAddr(&glob, (uint32_t*)glob.pBufRevAddr, (uint32_t)(offset >> 2), chunk);
          writeMemory(&glob, glob.pBufRevAddr, offset, chunk, dma);
          remaining -= chunk;
          offset += chunk;
          spin(&glob, TRUE);
        }

        /* Read back data from memory bank and check */
        remaining = glob.bankSize[i];
        offset = 0;
        location = 0;
        nErr = 0;
        while (remaining) {
          chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
          generateRevAddr(&glob, (uint32_t*)glob.pBufRevAddr, (uint32_t)(offset >> 2), chunk);
          verifyChunk(
            &glob,
            mask,
            glob.pBufRevAddr,
            width,
            location,
            chunk,
            &nErr,
            dma ? TRUE : FALSE,
            bRetryReads);
          location += (chunk / width);
          offset += chunk;
          remaining -= chunk;
          spin(&glob, FALSE);
        }
        if (nErr) {
          _tprintf(_T("*** %lld(0x%llX) error(s) encountered while verifying data\n"),
            (unsigned long long)nErr, (unsigned long long)nErr);
          bFailed = TRUE;
          break;
        }
      }

      _tprintf(_T(" Random data...\n"));
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
        if (!(glob.bankTestMask & (0x1U << i))) {
          continue;
        }
        if (bUseMask) {
          displayMask(glob.bankWidth[i], mask, maskStr + 14, ARRAY_LENGTH(maskStr) - 14);
        } else {
          maskStr[0] = _T('\0');
        }
        _tprintf(_T("   Bank %lu%s\r"), (unsigned long)i, maskStr);
        selectBank(&glob, i);
        width = glob.bankWidth[i];

        /* Write data to memory bank */
        remaining = glob.bankSize[i];
        offset = 0;
        seed(&glob.mtState, i + 1);
        while (remaining) {
          chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
          generateRandom(&glob.mtState, (uint32_t*)glob.pBufRandom, chunk);
          writeMemory(&glob, glob.pBufRandom, offset, chunk, dma);
          remaining -= chunk;
          offset += chunk;
          spin(&glob, TRUE);
        }

        /* Read back data from memory bank and check */
        remaining = glob.bankSize[i];
        offset = 0;
        location = 0;
        nErr = 0;
        seed(&glob.mtState, i + 1);
        while (remaining) {
          chunk = (GENERATE_CHUNK > remaining) ? (unsigned long)remaining : GENERATE_CHUNK;
          generateRandom(&glob.mtState, (uint32_t*)glob.pBufRandom, chunk);
          verifyChunk(
            &glob,
            mask,
            glob.pBufRandom,
            width,
            location,
            chunk,
            &nErr,
            dma ? TRUE : FALSE,
            bRetryReads);
          location += (chunk / width);
          offset += chunk;
          remaining -= chunk;
          spin(&glob, FALSE);
        }
        if (nErr) {
          _tprintf(_T("*** %lld(0x%llX) error(s) encountered while verifying data\n"),
            (unsigned long long)nErr, (unsigned long long)nErr);
          bFailed = TRUE;
          break;
        }
      }
      if (bFailed) {
        break;
      }
    }
    if (bFailed) {
      break;
    }
  }

  /***********************************************************************
  ** 20. Here we check that none of the sticky flags in STATUS indicate **
  **     that a DLL/DCM/PLL lost lock, or that a memory bank's          **
  **     "trained" status transitioned to zero.                         **
  ***********************************************************************/

  if (!checkStickyStatus(&glob, numLockFlags)) {
    ret = -1;
    goto done;
  }

  /***********************************************************************
  ** 21. Here we measure the throughput attainable between the host and **
  **     each memory bank (unless disabled on the command-line).        **
  ***********************************************************************/

  if (bMeasurePerf) {
    for (dma = 0; dma < 2; dma++) {
      if (!dma && !bUsePio) {
        continue;
      }
      if (dma && !bUseDma) {
        continue;
      }
      for (i = 0; i < cardInfo.NumRAMBank; i++) {
        const unsigned long chunk = GENERATE_CHUNK;
        double bytes, delta, rate;
        admxrc2_time_t start, expire, finish;

        if (!(glob.bankTestMask & (0x1U << i))) {
          continue;
        }

        _tprintf(_T("Measuring throughput for bank %lu using %s...\n"), (unsigned long)i, dma ? _T("DMA") : _T("programmed I/O"));
        selectBank(&glob, i);

        start = timeGet();
        expire = timeAdd(start, doubleToTime(2.0));
        bytes = 0.0;
        do {
          readMemory(&glob, glob.pBufChk, 0, chunk, dma);
          bytes += (float)chunk;
          finish = timeGet();
        } while (timeBefore(finish, expire));

        delta = timeToDouble(timeSubtract(finish, start));
        rate = bytes / delta;
        _tprintf(_T(" SSRAM to host throughput = %.1f MiB/s\n"), rate / 1048576.0f);

        start = timeGet();
        expire = timeAdd(start, doubleToTime(2.0));
        bytes = 0.0f;
        do {
          writeMemory(&glob, glob.pBufChk, 0, chunk, dma);
          bytes += (float)chunk;
          finish = timeGet();
        } while (timeBefore(finish, expire));

        delta = timeToDouble(timeSubtract(finish, start));
        rate = bytes / delta;
        _tprintf(_T(" Host to SSRAM throughput = %.1f MiB/s\n"), rate / 1048576.0f);
      }
    }
  }

  /***********************************************************************
  ** 22. Clean up.                                                      **
  ***********************************************************************/

done:
  if (glob.pBuf5a != NULL) {
    free(glob.pBuf5a);
  }

  if (glob.pBufOwnAddr != NULL) {
    free(glob.pBufOwnAddr);
  }

  if (glob.pBufRevAddr != NULL) {
    free(glob.pBufRevAddr);
  }

  if (glob.pBufRandom != NULL) {
    free(glob.pBufRandom);
  }

  if (glob.pBufChk != NULL) {
    free(glob.pBufChk);
  }

#if defined(_WIN32)
  if (glob.hDmaEvent != NULL) {
    CloseHandle(glob.hDmaEvent);
  }
#endif

  if (glob.hCard != ADMXRC2_HANDLE_INVALID_VALUE) {
    ADMXRC2_CloseCard(glob.hCard);
  }

  return ret;
}
