--
-- lclk_distribution.vhd - DCM for distributing LCLK in FPGA design in
--                         ADM-PCIE-6S1
--
-- Modules defined:
--
--    lclk_distribution    DCM_SP configured for x1 clock distribution with
--                         deskew.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library unisim;
use unisim.vcomponents.all;

entity lclk_distribution is
    generic(
        use_dcm          : in    boolean := true);
    port(
        lclk             : in    std_logic;
        rst_in           : in    std_logic;
        clk              : out   std_logic;
        rst_out          : out   std_logic);
end entity;

architecture mixed of lclk_distribution is

begin

    -- Use a DCM for distributing LCLK within the FPGA.
    distribute_lclk_dcm : if use_dcm generate
    
        signal logic0 : std_logic;
        signal lclk_ibufg : std_logic;
        signal lclk_bufg : std_logic;
        signal clk_unbuf : std_logic;
        signal clk_buf : std_logic;
        signal dcm_rst_count : std_logic_vector(2 downto 0) := (others => '0');
        signal dcm_rst : std_logic := '1';
        signal dcm_locked : std_logic;
        
    begin
    
        logic0 <= '0';
        
        ibufg_lclk : IBUFG
            port map(
                I => lclk,
                O => lclk_ibufg);
                
        bufg_lclk : BUFG
            port map(
                I => lclk_ibufg,
                O => lclk_bufg);
                
        gen_dcm_rst : process(rst_in, lclk_bufg)
        begin
            if rst_in = '1' then
                dcm_rst_count <= (others => '0');
                dcm_rst <= '1';
            elsif rising_edge(lclk_bufg) then
                if dcm_rst = '1' then
                    dcm_rst_count <= dcm_rst_count + 1;
                end if;
                if AND_reduce(dcm_rst_count) = '1' then
                    dcm_rst <= '0';
                end if;
            end if;
        end process;
        
        dcm_lclk : DCM_SP
            port map(
                RST => dcm_rst,
                CLKIN => lclk_ibufg,
                CLKFB => clk_buf,
                DSSEN => logic0,
                PSINCDEC => logic0,
                PSEN => logic0,
                PSCLK => lclk_ibufg,
                LOCKED => dcm_locked,
                STATUS => open,
                CLK0 => clk_unbuf,
                CLK90 => open,
                CLK180 => open,
                CLK270 => open,
                CLK2X => open,
                CLK2X180 => open,
                CLKFX => open,
                CLKFX180 => open,
                PSDONE => open);
                
        bufg_clk : BUFG
            port map(
                I => clk_unbuf,
                O => clk_buf);
                
        rst_out <= NOT dcm_locked;        
        clk <= clk_buf;
    end generate;
    
    -- Don't use a DCM for distributing LCLK within the FPGA.
    distribute_lclk_no_dcm : if NOT use_dcm generate
    begin
        clk <= lclk;
        rst_out <= rst_in;
    end generate;
    
end mixed;
