-- itest-pcie6s1.vhd - FPGA design demonstrating interrupt generation
--
-- (C) Copyright Alpha Data 2013
--
-- Example program for:
--
--   o ADM-PCIE-6S1
--
-- NOTES ON OPERATION
--
-- The logic in this FPGA design works in the way described
-- in the application note an-xrc06.pdf, in the doc/ directory
-- of this SDK.
--
-- 32 sources of interrupt are provided, each of which can be
-- triggered by writing a 1 to a corresponding bit of a write-
-- only register called TEST. When a source of interrupt has
-- is triggered, the corresponding bit of ISTAT becomes 1.
--
-- Each source of interrupt may be masked (disabled) by the
-- corresponding bit of IMASK being 1. The IMASK register resets
-- to 0xffffffff, ie. all interrupts disabled.
--
-- A write to byte 0 of the IARM register forces the FINTI_L signal
-- high (deasserted) for a single cycle, to "rearm" the FINTI_L signal.
-- The reason for including the IARM register is discussed in
-- an-xrc06.pdf .
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.localbus.all;

entity itest is
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic;
        lwrite           : in    std_logic;
        lads_l           : in    std_logic;
        lblast_l         : in    std_logic;
        lbterm_l         : inout std_logic;
        ld               : inout std_logic_vector(31 downto 0);
        la               : in    std_logic_vector(23 downto 2);
        lready_l         : inout std_logic;
        lbe_l            : in    std_logic_vector(3 downto 0);
        fholda           : in    std_logic;
        finti_l          : out   std_logic);
end itest;

architecture mixed of itest is

    signal clk : std_logic;
    signal rst : std_logic;
    signal lreset_i : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbterm_i : std_logic;
    signal lwrite_i : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lbe_iq_l : std_logic_vector(3 downto 0);
    signal ld_iq : std_logic_vector(31 downto 0);
    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_oe_l : std_logic;
    
    signal finti : std_logic;
    
    signal la_q : std_logic_vector(23 downto 2);
    signal fholda_q : std_logic;

    signal qlads : std_logic;
    signal ds_accept : std_logic;

    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;
    
    signal write_istat, write_test : std_logic;
    signal istat : std_logic_vector(31 downto 0);
    signal write_imask : std_logic;
    signal imask : std_logic_vector(31 downto 0);
    signal write_iarm : std_logic;
    signal iarm : std_logic;
    signal write_count : std_logic;
    signal count : std_logic_vector(31 downto 0);
    
    signal logic0, logic1 : std_logic;
    
    attribute IOB : string;
    -- Better performance from not mapping ld_o etc. into IOBs
    attribute IOB of ld_oe_l : signal is "FALSE";
    attribute IOB of ld_o : signal is "FALSE";
    
begin
    
    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --

    lads_i <= NOT lads_l;
    lblast_i <= NOT lblast_l;
    lbterm_i <= NOT lbterm_l;
    lreset_i <= NOT lreset_l;
    lwrite_i <= lwrite;
    
    --
    -- Register LD and LBE# (should be mapped into IOBs)
    --

    gen_lbe_iq_l : process(clk)
    begin
        if rising_edge(clk) then
            lbe_iq_l <= lbe_l;
        end if;
    end process;

    gen_ld_iq : process(clk)
    begin
        if rising_edge(clk) then
            ld_iq <= ld;
        end if;
    end process;

    --
    -- Generate a qualified version of 'lads_l', which is
    -- asserted when the FPGA is addressed AND the FPGA is
    -- not the local bus master.
    --

    qlads <= lads_i;
    ds_accept <= NOT la_q(23) AND NOT fholda_q;

    --
    -- Latch the local bus address on the 'lads_l' pulse.
    --

    latch_addr : process(clk)
    begin
        if rising_edge(clk) then
            la_q <= la;
            fholda_q <= fholda;
        end if;
    end process;
    
    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    
    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    lready_l <= lready_o_l when lready_oe_l = '0' else 'Z';
    
    --
    -- Drive the local data bus on a read.
    --

    gen_ld : for i in ld'range generate
        ld(i) <= ld_o(i) when ld_oe_l = '0' else 'Z';
    end generate;
    
    --
    -- Generate the 'ld_o' signal, outbound local bus data.
    --

    generate_ld_o : process(clk)
    begin
        if rising_edge(clk) then
            case la_q(4 downto 2) is
                when "000" =>
                    ld_o <= imask;
                when "001" =>
                    ld_o <= istat;
                when "100" =>
                    ld_o <= count;
                when others =>
                    ld_o <= (others => '-');
            end case;
        end if;
    end process;

    --
    -- generate 'finti_l' (FPGA interrupt)
    --

    gen_finti: process(clk, rst)
    begin
        if rst = '1' then
            finti <= '0';
        elsif rising_edge(clk) then
            if (not imask and istat) = EXT("0", 32) or iarm = '1' then
                finti <= '0';
            else
                finti <= '1';
            end if;
        end if;
    end process;

    finti_l <= not finti;

    --
    -- Implement the registers.
    --
    -- Register     Local addr      Type            Function
    -- -------------------------------------------------------------------
    -- IMASK        0x0             R/W             Interrupt mask
    -- ISTAT        0x4             R/W1C           Interrupt status
    -- IARM         0x8             WO/RAX          Interrupt rearm strobe
    -- TEST         0xC             WO              Sets ISTAT for testing
    -- COUNT        0x10            R/W             clk cycle counter
    --

    implement_registers : process(clk, rst)
        variable tmp : std_logic_vector(1 downto 0);
    begin
        if rst = '1' then
            write_imask <= '0';
            imask <= (others => '1');
            write_istat <= '0';
            write_test <= '0';
            istat <= (others => '0');
            write_iarm <= '0';
            iarm <= '0';
            write_count <= '0';
            count <= (others => '0');
        elsif rising_edge(clk) then
            if ds_xfer = '1' AND ds_write = '1' then
                if la_q(4 downto 2) = "000" then
                    write_imask <= '1';
                else
                    write_imask <= '0';
                end if;
                if la_q(4 downto 2) = "001" then
                    write_istat <= '1';
                else
                    write_istat <= '0';
                end if;
                if la_q(4 downto 2) = "010" then
                    write_iarm <= '1';
                else
                    write_iarm <= '0';
                end if;
                if la_q(4 downto 2) = "011" then
                    write_test <= '1';
                else
                    write_test <= '0';
                end if;
                if la_q(4 downto 2) = "100" then
                    write_count <= '1';
                else
                    write_count <= '0';
                end if;
            else
                write_imask <= '0';
                write_istat <= '0';
                write_iarm <= '0';
                write_test <= '0';
                write_count <= '0';
            end if;
            
            if write_imask = '1' then
                -- write to IMASK
                if lbe_iq_l(0) = '0' then
                    imask(7 downto 0) <= ld_iq(7 downto 0);
                end if;
                if lbe_iq_l(1) = '0' then
                    imask(15 downto 8) <= ld_iq(15 downto 8);
                end if;
                if lbe_iq_l(2) = '0' then
                    imask(23 downto 16) <= ld_iq(23 downto 16);
                end if;
                if lbe_iq_l(3) = '0' then
                    imask(31 downto 24) <= ld_iq(31 downto 24);
                end if;
            end if;
            
            tmp := write_istat & write_test;
            case tmp is
                when "00" | "11" =>
                    null;
                    
                when "10" =>
                    -- write to ISTAT - 1s clear corresponding bit
                    if lbe_iq_l(0) = '0' then
                        istat(7 downto 0) <= istat(7 downto 0) and not ld_iq(7 downto 0);
                    end if;
                    if lbe_iq_l(1) = '0' then
                        istat(15 downto 8) <= istat(15 downto 8) and not ld_iq(15 downto 8);
                    end if;
                    if lbe_iq_l(2) = '0' then
                        istat(23 downto 16) <= istat(23 downto 16) and not ld_iq(23 downto 16);
                    end if;
                    if lbe_iq_l(3) = '0' then
                        istat(31 downto 24) <= istat(31 downto 24) and not ld_iq(31 downto 24);
                    end if;
              
                when "01" =>
                    -- write to TEST    
                    if lbe_iq_l(0) = '0' then
                        istat(7 downto 0) <= istat(7 downto 0) or ld_iq(7 downto 0);
                    end if;
                    if lbe_iq_l(1) = '0' then
                        istat(15 downto 8) <= istat(15 downto 8) or ld_iq(15 downto 8);
                    end if;
                    if lbe_iq_l(2) = '0' then
                        istat(23 downto 16) <= istat(23 downto 16) or ld_iq(23 downto 16);
                    end if;
                    if lbe_iq_l(3) = '0' then
                        istat(31 downto 24) <= istat(31 downto 24) or ld_iq(31 downto 24);
                    end if;
                    
                when others =>
                    null;
            end case;
            
            -- write to IARM
            if write_iarm = '1' then
                if lbe_iq_l(0) = '0' then
                    iarm <= '1';
                end if;
            else
                iarm <= '0';
            end if;

            -- Cycle counter
            if istat(0) = '0' then
                -- Allow writes to cycle counter when ISTAT[0] is clear.
                if write_count = '1' then
                    -- write to COUNT
                    if lbe_iq_l(0) = '0' then
                        count(7 downto 0) <= ld_iq(7 downto 0);
                    end if;
                    if lbe_iq_l(1) = '0' then
                        count(15 downto 8) <= ld_iq(15 downto 8);
                    end if;
                    if lbe_iq_l(2) = '0' then
                        count(23 downto 16) <= ld_iq(23 downto 16);
                    end if;
                    if lbe_iq_l(3) = '0' then
                        count(31 downto 24) <= ld_iq(31 downto 24);
                    end if;
                end if;
            else
                -- Increment the cycle counter when ISTAT[0] is 1.
                count <= count + 1;
            end if;
        end if;
    end process;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0 : plxdssm
        generic map(
            -- Better performance from not mapping local bus outputs into IOBs
            output_fast => false)
        port map(
            clk              => clk,
            rst              => rst,
            sr               => logic0,
            qlads            => qlads,
            lblast           => lblast_i,
            lbterm           => lbterm_i,
            lwrite           => lwrite_i,
            eld_oe           => open,
            ld_oe_l          => ld_oe_l,
            lready_o_l       => lready_o_l,
            lready_oe_l      => lready_oe_l,
            lbterm_o_l       => lbterm_o_l,
            lbterm_oe_l      => lbterm_oe_l,
            idle             => open,
            transfer         => ds_xfer,
            decode           => ds_decode,
            write            => ds_write,
            accept           => ds_accept,
            ready            => logic1,
            stop             => logic1);
    
    --
    -- Use a DCM for distributing LCLK within the FPGA.
    --
    
    lclk_dist0 : lclk_distribution
        port map(
            lclk => lclk,
            rst_in => lreset_i,
            clk => clk,
            rst_out => rst);
            
end mixed;
