--
-- master-pcie6s1.vhd - Direct Master example FPGA design
--
-- (C) Copyright Alpha Data 2013
--
-- Example program for:
--
--   o ADM-PCIE-6S1
--
-- NOTES ON OPERATION
-- ==================
--
-- LA[23:2] are connected to the FPGA.
--
-- This design implements the following registers, accessible by
-- direct slave transfers:
--
--   Offset Type  Name     Function
--   ===================================================================
--
--   0x0    WO    ADDR     Local bus address for direct master cycles.
--
--   0x4    WO    WDATA    Data to be written in next direct master
--                         write cycle.
--
--   0x8    WO    CFG      Selects what type of direct master cycle to
--                         perform; bit fields as follows:
--
--                         [0] WRITE   1 => perform a write
--                                     0 => perform a read
--                         [1] PLXS0   1 => assert PLXS0# during direct master cycle
--                                     0 => don't assert PLXS0# during direct master cycle
--
--   0xC    RO    RDATA    Data read in last direct master read cycle.
--
--   0x10   RO    STAT     Status register; bit fields as follows:
--
--                         [0] IDLE    1 => no direct master cycle
--                                          in progress
--                                     0 => direct master cycle in
--                                          progress
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;

library work;
use work.localbus.all;

entity master is
    generic(
        la_top           : in    natural := 23);
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic;
        lwrite           : inout std_logic;
        lads_l           : inout std_logic;
        lblast_l         : inout std_logic;
        lbterm_l         : inout std_logic;
        ld               : inout std_logic_vector(31 downto 0);
        la               : inout std_logic_vector(la_top downto 2);
        lready_l         : inout std_logic; 
        lbe_l            : inout std_logic_vector(3 downto 0);
        llocki_l         : out   std_logic;
        plxs0_l          : out   std_logic;
        fhold            : out   std_logic; 
        fholda           : in    std_logic);
end master;

architecture mixed of master is
    
    constant DM_WINDOW_LOCAL_BASE : std_logic_vector(31 downto 0) := X"FF000000";
    constant DM_WINDOW_LOCAL_MASK : std_logic_vector(31 downto 0) := X"007FFFFC";
    
    signal clk : std_logic;
    signal rst : std_logic;
    signal lads_i : std_logic;
    signal lblast_i : std_logic;
    signal lbterm_i : std_logic;
    signal lready_i : std_logic;
    signal lreset_i : std_logic;
    signal lwrite_i : std_logic;
    signal ld_iq : std_logic_vector(31 downto 0);
    signal lads_o_l : std_logic;
    signal lads_oe_l : std_logic;
    signal lbe_iq_l : std_logic_vector(3 downto 0);
    signal lbe_o_l : std_logic_vector(3 downto 0);
    signal lbe_oe_l : std_logic;
    signal lblast_o_l : std_logic;
    signal lblast_oe_l : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lwrite_o : std_logic;
    signal lwrite_oe_l : std_logic;
    
    signal la_oe_l : std_logic;
    signal la_o : std_logic_vector(la_top downto 2);
    signal ld_oe_l : std_logic_vector(31 downto 0);
    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_o_ds : std_logic_vector(31 downto 0);
    signal ld_o_dm : std_logic_vector(31 downto 0);
    signal ccs_oe_l : std_logic;
    signal ccs_o_l : std_logic;
    
    signal la_q : std_logic_vector(la_top downto 2);
    signal fholda_q : std_logic;
    
    signal ds_qlads : std_logic;
    signal ds_accept : std_logic;
    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;
    signal ds_write : std_logic;
    signal ds_eld_oe : std_logic;
 
    signal write_addr : std_logic; 
    signal reg_addr : std_logic_vector(31 downto 0);
    signal write_wdata : std_logic; 
    signal reg_wdata : std_logic_vector(31 downto 0);
    signal write_cfg : std_logic; 
    signal reg_cfg : std_logic_vector(31 downto 0);
    signal reg_rdata : std_logic_vector(31 downto 0);
    signal reg_stat : std_logic_vector(31 downto 0);

    signal dm_req : std_logic;
    signal dm_xfer : std_logic;
    signal dm_eld_oe : std_logic;
    signal dm_done : std_logic;

    signal logic0, logic1 : std_logic;
    
    attribute IOB : string;
    -- Better performance from not mapping ld_o etc. into IOBs
    attribute IOB of ld_oe_l : signal is "FALSE";
    attribute IOB of ld_o : signal is "FALSE";
    -- Better performance from not mapping la_o, la_oe_l and lbe_o_l into IOBs
    attribute IOB of la_o : signal is "FALSE";
    attribute IOB of la_oe_l : signal is "FALSE";
    attribute IOB of lbe_oe_l : signal is "FALSE";
    
begin
    
    -- -------------------------------------------------------------------
    -- Common
    -- -------------------------------------------------------------------

    logic0 <= '0';
    logic1 <= '1';

    --    
    -- Convert inputs to active high.
    --

    lreset_i <= NOT lreset_l;
    lready_i <= NOT lready_l;
    lbterm_i <= NOT lbterm_l;

    --
    -- Register local bus data on the way in.
    --

    register_ld_iq: process(clk)
    begin
        if rising_edge(clk) then
            ld_iq <= ld;
        end if;
    end process;

    --
    -- Generate 'ld_o', the data for the LD pins, depending on whether the
    -- current local bus transfer a direct slave transfer or a direct master
    -- transfer.
    --

    generate_ld_o : process(rst, clk)
    begin
        if rising_edge(clk) then
            if ds_eld_oe = '1' then
                ld_o <= ld_o_ds;
            else
                ld_o <= ld_o_dm;
            end if;
        end if;
    end process;

    --
    -- Drive the local data bus when responding to a direct slave read cycle
    -- or direct master write cycle.
    --

    gen_ld_oe_l : process(clk, rst)
    begin
        if rst = '1' then
            ld_oe_l <= (others => '1');
        elsif rising_edge(clk) then
            ld_oe_l <= (others => not (ds_eld_oe or dm_eld_oe));
        end if;
    end process;

    gen_ld : for i in ld'range generate
        ld(i) <= ld_o(i) when ld_oe_l(i) = '0' else 'Z';
    end generate;

    --
    -- Use a DCM for distributing LCLK within the FPGA.
    --
    
    lclk_dist0 : lclk_distribution
        port map(
            lclk => lclk,
            rst_in => lreset_i,
            clk => clk,
            rst_out => rst);
            
    -- -------------------------------------------------------------------
    -- Direct Slave
    -- -------------------------------------------------------------------

    --
    -- Convert inputs to active high.
    --

    lads_i <= NOT lads_l;
    lblast_i <= NOT lblast_l;
    lbterm_i <= lbterm_i;
    lwrite_i <= lwrite;

    --
    -- Register local bus byte enables on the way in.
    --

    register_lbe_iq_l : process(clk)
    begin
        if rising_edge(clk) then
            lbe_iq_l <= lbe_l;
        end if;
    end process;

    --
    -- 'ds_qlads' is 'lads_l' qualified as a direct slave access to the FPGA.
    --

    ds_qlads <= lads_i; 
    ds_accept <= NOT la_q(23) AND NOT fholda_q;
    
    --
    -- Latch local bus address on direct slave transfers.
    --

    register_la : process(clk)
    begin
        if rising_edge(clk) then
            la_q <= la;
            fholda_q <= fholda;
        end if;
    end process;

    --
    -- Drive the bus when responding to a direct slave cycle.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    lready_l <= lready_o_l when lready_oe_l = '0' else 'Z';

    --
    -- Implement the registers that are writable via direct slave cycles.
    --

    implement_registers : process(clk, rst)
        variable tmp32 : std_logic_vector(31 downto 0);
    begin
        if rst = '1' then
            write_addr <= '0';
            dm_req <= '0';
            reg_addr <= (others => '0');
            write_wdata <= '0';
            reg_wdata <= (others => '0');
            write_cfg <= '0';
            reg_cfg <= (others => '0');
        elsif rising_edge(clk) then
            if ds_xfer = '1' AND ds_write = '1' AND la_q(4 downto 2) = "000" then
                write_addr <= '1';
            else
                write_addr <= '0';
            end if;
            
            if ds_xfer = '1' AND ds_write = '1' AND la_q(4 downto 2) = "001" then
                write_wdata <= '1';
            else
                write_wdata <= '0';
            end if;
            
            if ds_xfer = '1' AND ds_write = '1' AND la_q(4 downto 2) = "010" then
                write_cfg <= '1';
            else
                write_cfg <= '0';
            end if;
            
            if write_addr = '1' then
                -- Write to ADDR register
                if lbe_iq_l(0) = '0' then
                    reg_addr(7 downto 2) <= ld_iq(7 downto 2);
                end if;
                if lbe_iq_l(1) = '0' then
                    reg_addr(15 downto 8) <= ld_iq(15 downto 8);
                end if;
                if lbe_iq_l(2) = '0' then
                    reg_addr(23 downto 16) <= ld_iq(23 downto 16);
                end if;
                if lbe_iq_l(3) = '0' then
                    reg_addr(31 downto 24) <= ld_iq(31 downto 24);
                end if;
            end if;

            --
            -- Writing to byte 3 of ADDR register also triggers a
            -- direct master cycle.
            --
            if write_addr = '1' AND lbe_iq_l(3) = '0' then
                dm_req <= '1';
            else
                dm_req <= '0';
            end if;
            
            if write_wdata = '1' then
                -- Write to WDATA register.
                if lbe_iq_l(0) = '0' then
                    reg_wdata(7 downto 0) <= ld_iq(7 downto 0);
                end if;
                if lbe_iq_l(1) = '0' then
                    reg_wdata(15 downto 8) <= ld_iq(15 downto 8);
                end if;
                if lbe_iq_l(2) = '0' then
                    reg_wdata(23 downto 16) <= ld_iq(23 downto 16);
                end if;
                if lbe_iq_l(3) = '0' then
                    reg_wdata(31 downto 24) <= ld_iq(31 downto 24);
                end if;
            end if;
            
            if write_cfg = '1' then
                -- Write to CFG register
                if lbe_iq_l(0) = '0' then
                    reg_cfg(1 downto 0) <= ld_iq(1 downto 0);
                end if;
            end if;
        end if;
    end process;

    --
    -- Implement the registers that are readable via direct slave cycles.
    --

    implment_reg_rdata : process(rst, clk)
    begin
        if rst = '1' then
            reg_rdata <= X"DEADBEEF";
        elsif rising_edge(clk) then
            if dm_xfer = '1' then
                reg_rdata <= ld_iq;
            end if;
        end if;
    end process;

    reg_stat(0)           <= dm_done;
    reg_stat(31 downto 1) <= (others => '0');

    --
    -- Generate 'ld_o_ds', the data for the LD pins when current bus cycle
    -- is a direct slave transfer.
    --

    generate_ld_o_ds : process(
        la_q,
        reg_addr,
        reg_wdata,
        reg_cfg,
        reg_rdata,
        reg_stat)
    begin
        case la_q(4 downto 2) is
            when "000" =>
                ld_o_ds <= reg_addr;
            when "001" =>
                ld_o_ds <= reg_wdata;
            when "010" =>
                ld_o_ds <= reg_cfg;
            when "011" =>
                ld_o_ds <= reg_rdata;
            when "100" =>
                ld_o_ds <= reg_stat;
            when others =>
                ld_o_ds <= (others => '-');
        end case;
    end process;

    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    plxdssm0 : plxdssm
        port map(
            clk => clk,
            rst => rst,
            sr => logic0,
            qlads => ds_qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => ds_eld_oe,
            ld_oe_l => open,
            lready_o_l => lready_o_l,
            lready_oe_l => lready_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => open,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            accept => ds_accept,
            ready => logic1,
            stop => logic1); 

    -- -------------------------------------------------------------------
    -- Direct Master
    -- -------------------------------------------------------------------

    --
    -- Drive the bus when performing a direct master cycle.
    --

    gen_la_oe_l : process(rst, clk)
    begin
        if rst = '1' then
            la_oe_l <= '1';
            lbe_oe_l <= '1';
        elsif rising_edge(clk) then
            -- Can drive LA and LBE# whenever arbiter grants us the local bus
            la_oe_l <= NOT fholda;
            lbe_oe_l <= NOT fholda;
        end if;
    end process;

    plxs0_l <= ccs_o_l when ccs_oe_l = '0' else 'Z';
    la <= la_o when la_oe_l = '0' else (others => 'Z');
    lbe_l <= lbe_o_l when lbe_oe_l = '0' else (others => 'Z');
    lads_l <= lads_o_l when lads_oe_l = '0' else 'Z';
    lblast_l <= lblast_o_l when lblast_oe_l = '0' else 'Z';
    lwrite <= lwrite_o when lwrite_oe_l = '0' else 'Z';
    llocki_l <= '1';

    --
    -- Generate the address and byte enables to be output onto the local bus.
    --

    generate_la_o : process(clk)
      variable tmp32 : std_logic_vector(31 downto 0);
    begin
        if rising_edge(clk) then
            tmp32 := (reg_addr AND DM_WINDOW_LOCAL_MASK) OR DM_WINDOW_LOCAL_BASE;
            la_o <= tmp32(la_top downto 2);
        end if;
    end process;
    
    lbe_o_l <= "0000";

    --
    -- There is currently only one source of data for direct master writes.
    --

    ld_o_dm <= reg_wdata;

    --
    -- Instantiate the direct master state machine; performs local bus
    -- cycles as a master.
    --

    plxdmsm0 : plxdmsm
        generic map(
            -- Better performance from not mapping local bus outputs into IOBs
            output_fast => false)
        port map(
            clk => clk,
            rst => rst,
            sr => logic0,
            ccs => reg_cfg(1),
            request => dm_req,
            stop => logic1,
            transfer => dm_xfer,
            done => dm_done,
            write => reg_cfg(0),
            lready => lready_i,
            lbterm => lbterm_i,
            fholda => fholda,
            fhold => fhold,
            la_oe_l => open,
            lads_o_l => lads_o_l,
            lads_oe_l => lads_oe_l,
            ccs_o_l => ccs_o_l,
            ccs_oe_l => ccs_oe_l,
            lbe_oe_l => open,
            lblast_o_l => lblast_o_l,
            lblast_oe_l => lblast_oe_l,
            eld_oe => dm_eld_oe,
            ld_oe_l => open,
            lwrite => lwrite_o,
            lwrite_oe_l => lwrite_oe_l);

end mixed;
