--
-- simple-pcie6s1.vhd - Trivial design demonstrating how to implement
--                      registers in the FPGA.
--
-- (C) Copyright Alpha Data 2013
--
-- Example program for:
--
--   o ADM-PCIE-6S1
--

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.localbus.all;

entity simple is
    port(
        lclk             : in    std_logic;
        lreset_l         : in    std_logic; 
        lwrite           : in    std_logic;
        lads_l           : in    std_logic;
        lblast_l         : in    std_logic;
        lbterm_l         : inout std_logic;
        la               : in    std_logic_vector(23 downto 2);
        ld               : inout std_logic_vector(31 downto 0);
        lready_l         : out   std_logic; 
        lbe_l            : in    std_logic_vector(3 downto 0);
        fholda           : in    std_logic);
end simple;

architecture mixed of simple is
    
    signal clk : std_logic;
    signal rst : std_logic;
    signal lreset_i : std_logic;
    signal lblast_i : std_logic;
    signal lads_i : std_logic;
    signal lwrite_i : std_logic;
    signal lready_o_l : std_logic;
    signal lready_oe_l : std_logic;
    signal lbterm_i : std_logic;
    signal lbterm_o_l : std_logic;
    signal lbterm_oe_l : std_logic;
    signal ld_o : std_logic_vector(31 downto 0);
    signal ld_iq : std_logic_vector(31 downto 0);
    signal ld_oe_l : std_logic;
    signal lbe_iq_l : std_logic_vector(3 downto 0);

    signal qlads : std_logic;
    signal ds_accept : std_logic;
    signal ds_xfer : std_logic;
    signal ds_decode : std_logic;    
    signal ds_write : std_logic;

    signal la_q : std_logic_vector(23 downto 2);
    signal fholda_q : std_logic;
    
    signal write_reg : std_logic_vector(1 downto 0);
    signal reg0 : std_logic_vector(31 downto 0);
    signal reg0_rev : std_logic_vector(31 downto 0);
    signal reg1 : std_logic_vector(31 downto 0);
    
    signal logic0, logic1 : std_logic;
    
    attribute IOB : string;
    -- Better performance from not mapping ld_o etc. into IOBs
    attribute IOB of ld_oe_l : signal is "FALSE";
    attribute IOB of ld_o : signal is "FALSE";
    
begin

    logic0 <= '0';
    logic1 <= '1';

    --
    -- Convert the inputs to active high.
    --

    lads_i <= NOT lads_l;
    lblast_i <= NOT lblast_l;
    lbterm_i <= NOT lbterm_l;
    lreset_i <= NOT lreset_l;
    lwrite_i <= lwrite;
    
    --
    -- Register 'ld' and 'lbe_l' (should be absorbed into IOBs)
    --

    register_ld_lbe : process(clk)
    begin
        if rising_edge(clk) then
            ld_iq <= ld;
            lbe_iq_l <= lbe_l;
        end if;
    end process;
    
    --
    -- Generate a qualified version of 'lads_l', which is
    -- asserted when the FPGA is addressed AND the FPGA is
    -- not the local bus master.
    --

    qlads <= lads_i;
    ds_accept <= NOT la_q(23) AND NOT fholda_q;

    --
    -- Latch the local bus address on the 'lads_l' pulse.
    --

    latch_addr : process(clk)
    begin
        if rising_edge(clk) then
            la_q <= la;
        end if;
    end process;
    
    --
    -- 'lbterm_l' should only be driven when the FPGA is addressed; otherwise
    -- float, because the control logic on the card might also drive it.
    --

    lbterm_l <= lbterm_o_l when lbterm_oe_l = '0' else 'Z';
    
    --
    -- 'lready_l' should only be driven when the FPGA is addressed; otherwise
    -- float because the control logic on the card might also drive it.
    --

    lready_l <= lready_o_l when lready_oe_l = '0' else 'Z';
    
    --
    -- Drive the local data bus on a read.
    --

    ld <= ld_o when ld_oe_l = '0' else (others => 'Z');
    
    --
    -- If the current cycle is a write, update the registers
    --

    update_regs : process(clk, rst)
    begin
        if rst = '1' then
            write_reg <= (others => '0');
            reg0 <= (others => '0');
            reg1 <= (others => '0');
        elsif rising_edge(clk) then
            write_reg(0) <= ds_xfer AND ds_write AND NOT la_q(2);
            write_reg(1) <= ds_xfer AND ds_write AND la_q(2);
            
            if write_reg(0) = '1' then
                if lbe_iq_l(0) = '0' then
                    reg0(7 downto 0) <= ld_iq(7 downto 0);
                end if;
                if lbe_iq_l(1) = '0' then
                    reg0(15 downto 8) <= ld_iq(15 downto 8);
                end if;
                if lbe_iq_l(2) = '0' then
                    reg0(23 downto 16) <= ld_iq(23 downto 16);
                end if;
                if lbe_iq_l(3) = '0' then
                    reg0(31 downto 24) <= ld_iq(31 downto 24);
                end if;
            end if;
            
            if write_reg(1) = '1' then
                if lbe_iq_l(0) = '0' then
                    reg1(7 downto 0) <= ld_iq(7 downto 0);
                end if;
                if lbe_iq_l(1) = '0' then
                    reg1(15 downto 8) <= ld_iq(15 downto 8);
                end if;
                if lbe_iq_l(2) = '0' then
                    reg1(23 downto 16) <= ld_iq(23 downto 16);
                end if;
                if lbe_iq_l(3) = '0' then
                    reg1(31 downto 24) <= ld_iq(31 downto 24);
                end if;
            end if;
        end if;
    end process;

    reg0_rev <= reg0(3 downto 0) &
                reg0(7 downto 4) &
                reg0(11 downto 8) &
                reg0(15 downto 12) &
                reg0(19 downto 16) &
                reg0(23 downto 20) &
                reg0(27 downto 24) &
                reg0(31 downto 28);

    --
    -- Generate the 'ld_o' signal, outbound data for local bus
    --

    generate_ld_o : process(clk)
    begin
        if rising_edge(clk) then
            if la_q(2) = '1' then
                ld_o <= reg1;
            else
                ld_o <= reg0_rev;
            end if;
        end if;
    end process;
        
    --
    -- Instantiate the direct slave state machine; monitors the local bus for
    -- direct slave cycles and responds appropriately.
    --

    dssm : plxdssm
        generic map(
            -- Better performance from not mapping local bus outputs into IOBs
            output_fast => false)
        port map(
            clk => clk,
            rst => rst,
            sr => logic0,
            qlads => qlads,
            lblast => lblast_i,
            lbterm => lbterm_i,
            lwrite => lwrite_i,
            eld_oe => open,
            ld_oe_l => ld_oe_l,
            lready_o_l  => lready_o_l,
            lready_oe_l => lready_oe_l,
            lbterm_o_l => lbterm_o_l,
            lbterm_oe_l => lbterm_oe_l,
            idle => open,
            transfer => ds_xfer,
            decode => ds_decode,
            write => ds_write,
            accept => ds_accept,
            ready => logic1,
            stop => logic1);

    --
    -- Use a DCM for distributing LCLK within the FPGA.
    --
    
    lclk_dist0 : lclk_distribution
        port map(
            lclk => lclk,
            rst_in => lreset_i,
            clk => clk,
            rst_out => rst);
            
end mixed;
